import polars as pl
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np


def do_plot(file_path: str, position=3.5):
    
    # # 设置全局字体样式
    # plt.rcParams.update({
    #     'xtick.labelsize': 12,
    #     'ytick.labelsize': 12,
    #     'xtick.labelweight': 'bold',
    #     'ytick.labelweight': 'bold',
    #     'axes.labelweight': 'bold'
    # })
    
    fig = plt.figure(figsize=(20, 10))
    axes = fig.add_subplot(1, 1, 1)
    axes.tick_params(axis='both', width=4, labelsize=16)
    axes.grid(visible=True)
    data = pl.read_csv(file_path, separator=",")
    data = data.to_pandas()
    sns.lineplot(data, x="idx", y="value", hue="tag", ax=axes, marker="o", markersize=10,   markeredgewidth=1.5,  linewidth=2 )
    xmin = data["idx"].min()
    xmax = data["idx"].max()
    
    axes.set_ylabel("PhreQ", fontdict={"fontsize": 14})
    axes.set_xlabel("Iter", fontdict={"fontsize": 14})

    axes.set_xticks(np.arange(xmin-1, xmax + 2, 1))  # +1确保包含最大值
    axes.set_xlim(xmin=xmin-1, xmax=xmax+1)  # 设置x轴范围

    axes.axvline(x=position, color='red', linestyle='--', linewidth=1)

    y_min, y_max = axes.get_ylim()  # 获取y轴范围元组
    axes.text(position, y_max*0.95, 'DatasetChange', fontdict={"fontsize": 14, "fontweight": "bold"},
              rotation=0, verticalalignment='top',
              bbox=dict(facecolor='white', alpha=0.8))

    plt.savefig("{}.png".format(file_path))
    


def plot_time(file_path: str):
    fig = plt.figure(figsize=(20, 10))
    axes = fig.add_subplot(1, 1, 1)
    axes.tick_params(axis='both', width=4, labelsize=16)
    
    axes.grid(visible=True)
    data = pl.read_csv(file_path, separator=",")
    data = data.to_pandas()
    xmin = data["idx"].min()
    xmax = data["idx"].max()

    axes.set_xticks(np.arange(xmin-1, xmax + 2, 1))  # +1确保包含最大值
    axes.set_xlim(xmin=xmin-1, xmax=xmax+1)  # 设置x轴范围
    axes.set_ylabel("Time(min)")
    axes.set_xlabel("Iter")
    sns.lineplot(data, x="idx", y="time", ax=axes, marker="o", markersize=10,   markeredgewidth=1.5,  linewidth=2)
    fig.savefig("{}.png".format(file_path))
    pass


def main():
    do_plot("/root/projects/gsda/gseda/src/gseda/plotter/smc_iter_data.csv", position=3.5)
    do_plot("gseda/src/gseda/plotter/icing_iter_data.csv", position=6.5)
    plot_time("/root/projects/gsda/gseda/src/gseda/plotter/speed_data.csv")
    pass


if __name__ == "__main__":
    main()



import pysam
import multiprocessing as mp
import pathlib
from typing import Set
from tqdm import tqdm
import subprocess


def extract_query_name_from_bam(bam_path):
    smc_names = set()
    with pysam.AlignmentFile(filename=bam_path, mode="rb", threads=mp.cpu_count()) as bam_in:
        for record in bam_in.fetch(until_eof=True):
            smc_names.add(record.query_name)
    return smc_names


def dump_subset_smc_bam(smc_bam_path: str, smc_names: Set[str]) -> str:
    p = pathlib.Path(smc_bam_path)

    subset_bam_path = p.parent.joinpath(f"{p.stem}.subset.bam")
    with pysam.AlignmentFile(filename=smc_bam_path, mode="rb", threads=mp.cpu_count() // 2, check_sq=False) as bam_in:
        with pysam.AlignmentFile(filename=str(subset_bam_path), mode="wb", threads=mp.cpu_count() // 2, check_sq=False, header=bam_in.header) as bam_out:
            for record in tqdm(bam_in.fetch(until_eof=True), desc=f"dumping {smc_bam_path}"):
                if record.query_name in smc_names:
                    bam_out.write(record)
    return str(subset_bam_path)


def run_asrtc(ref_fa, subreads_bam, smc_bam, prefix):
    cmd = f"asrtc --ref-fa {ref_fa} -q {subreads_bam} -t {smc_bam} --np-range 0:100 --rq-range 0.99:1.1 -p {prefix} -m 2 -M 4 -o 4,24 -e 2,1"
    print(f"cmd:{cmd}")
    subprocess.check_call(cmd, shell=True)


def main():

    bam_path = "/data1/ccs_data/saisuofei/20251204_241201Y0002_Run0001/Group_0/barcodes_reads_cons_gen_amplicon/Consensus/Bam/Group_0_Adaptor-barcode203-0.sort.bam"
    smc_bam_path = "/data1/ccs_data/saisuofei/20251204_241201Y0002_Run0001.smc_all_reads.bam"
    subreads_bam_path = "/data1/ccs_data/saisuofei/20251204_241201Y0002_Run0001_adapter.bam"
    ref_fasta = "/data1/ccs_data/saisuofei/Group_0_Adaptor-barcode203-0/Group_0_Adaptor-barcode203-0/ref.fasta"
    smc_names = extract_query_name_from_bam(bam_path=bam_path)
    print(f"num smc reads:{len(smc_names)}")
    smc_subset_bam_path = dump_subset_smc_bam(smc_bam_path, smc_names)
    run_asrtc(ref_fa=ref_fasta, subreads_bam=subreads_bam_path,
              smc_bam=smc_subset_bam_path, prefix="saisuofei-barcode-203")

    pass


if __name__ == "__main__":
    main()
    pass



import pysam
from tqdm import tqdm
import numpy as np
from typing import List
import seaborn as sns
import matplotlib.pyplot as plt
import pathlib


class ChannelLength:
    def __init__(self, ch):
        self.ch = ch
        self.lengths = []

    def median(self):
        return np.median(np.array(self.lengths))


def read_bam_info(bam_file: str):
    result = []

    with pysam.AlignmentFile(
        filename=bam_file, mode="rb", check_sq=False, threads=40
    ) as reader:
        for record in tqdm(reader.fetch(until_eof=True), desc=f"reading {bam_file}"):
            ch = record.get_tag("ch")
            cx = record.get_tag("cx")
            if cx < 3:
                continue

            if len(result) == 0:
                result.append(ChannelLength(ch=ch))
            
            if result[-1].ch != ch:
                # if len(result) > 10000:
                #     break
                result.append(ChannelLength(ch=ch))

            assert result[-1].ch == ch, "{}!={}".format(result[-1].ch, ch)
            result[-1].lengths.append(record.query_length)

    return result


def plot(result: List[ChannelLength], o_path: str):
    result = [r for r in result if len(r.lengths) >= 3]
    medians = [r.median() for r in result]
    # print(medians)
    x = list(range(len(medians)))

    # 设置 seaborn 风格
    sns.set(style="whitegrid")

    # 创建折线图
    plt.figure(figsize=(40, 21))
    sns.lineplot(x=x, y=medians, marker="o")

    plt.xlabel("Channel Index")
    plt.ylabel("Median Read Length")
    plt.title("Median Read Length per Channel")

    # 保存图片
    plt.tight_layout()
    plt.savefig(o_path)
    print(f"Saved plot to {o_path}")
    plt.close()
    pass


def main_cli():
    bam = "/data/ccs_data/speed-test/4.6k5h/20250424_240601Y0005_Run0001_adapter.bam"
    o_path = "{}.incomming_lengths.png".format(pathlib.Path(bam).stem)
    incomming_lengths = read_bam_info(bam)
    plot(incomming_lengths, o_path=o_path)
    
    pass

if __name__ == "__main__":
    main_cli()


import pysam
from tqdm import tqdm
import numpy as np
from typing import List
import argparse
from numba import njit


@njit
def sliding_topn_square_sum_numba(arr: np.ndarray, k: int, topn: int) -> np.ndarray:
    n = len(arr)
    output = np.empty(n - k + 1)

    for i in range(n - k + 1):
        window = arr[i : i + k]
        # 获取 topN（不排序），使用 np.partition
        idx = np.argpartition(window, -topn)[-topn:]
        square_sum = 0.0
        for j in range(topn):
            val = window[idx[j]]
            square_sum += val * val
        output[i] = square_sum

    return output


class ChannelLength:
    def __init__(self, ch):
        self.ch = ch
        self.lengths = []

    def median(self):
        return float(np.median(np.array(self.lengths)))


def read_channel_lengths(bam_file: str):
    result = []

    with pysam.AlignmentFile(
        filename=bam_file, mode="rb", check_sq=False, threads=40
    ) as reader:
        for record in tqdm(reader.fetch(until_eof=True), desc=f"reading {bam_file}"):
            ch = record.get_tag("ch")
            cx = record.get_tag("cx")
            if cx < 3:
                continue

            if len(result) == 0:
                result.append(ChannelLength(ch=ch))

            if result[-1].ch != ch:
                result.append(ChannelLength(ch=ch))

            assert result[-1].ch == ch, "{}!={}".format(result[-1].ch, ch)
            result[-1].lengths.append(record.query_length)

    return result


def mem_est(
    channel_lengths: List[ChannelLength], min_passes: int, window: int, top_n: int
):
    result = [r for r in channel_lengths if len(r.lengths) >= min_passes]
    medians = [r.median() for r in result]
    medians = np.array(medians)

    mems = sliding_topn_square_sum_numba(medians, window, top_n)
    # 4 for int. 2 for each strand
    print("PeakUsed EST. {:.2f}GB".format(np.max(mems) * 4 * 2 / (1024 * 1024 * 1024)))

def main_cli():
    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("sbr_bam", type=str)
    parser.add_argument("--min-passes", required=True, type=int, dest="min_passes")
    parser.add_argument("--win", required=True, type=int)
    parser.add_argument("--topn", required=True, type=int)
    
    args = parser.parse_args()
    channel_lengths = read_channel_lengths(bam_file=args.sbr_bam)
    mem_est(channel_lengths=channel_lengths, min_passes=args.min_passes, window=args.win, top_n=args.topn)
        
    pass


if __name__ == "__main__":
    main_cli()



import pysam
from tqdm import tqdm
from argparse import Namespace
import argparse


def read_wanted_channels(filename: str):

    channels = set()
    if (
        filename.endswith("fa")
        or filename.endswith("fasta")
        or filename.endswith("fna")
        or filename.endswith("fq")
        or filename.endswith("fastq")
    ):
        with pysam.FastxFile(filename=filename) as fh:
            for entry in fh:
                ch = entry.name.split("/")[1]
                try:
                    ch = int(ch)
                except:
                    continue
                channels.add(ch)
    elif filename.endswith("bam"):
        with pysam.AlignmentFile(
            filename=filename, mode="rb", threads=40, check_sq=False
        ) as bam_h:
            for record in bam_h.fetch(until_eof=True):
                channels.add(int(record.get_tag("ch")))

    return channels


def dump_bam_according_to_fastx_channel(subreads_bam: str, fastx_file: str, infix: str):
    channels = read_wanted_channels(fastx_file)
    o_bam_filename = "{}.{}.bam".format(subreads_bam.rsplit(".", maxsplit=1)[0], infix)
    with pysam.AlignmentFile(
        subreads_bam, mode="rb", threads=40, check_sq=False
    ) as in_bam:
        with pysam.AlignmentFile(
            o_bam_filename, mode="wb", threads=40, check_sq=False, header=in_bam.header
        ) as out_bam:

            for record in tqdm(
                in_bam.fetch(until_eof=True), desc=f"dumping {o_bam_filename}"
            ):
                ch = int(record.get_tag("ch"))
                if ch in channels:
                    out_bam.write(record)


def main(args):

    dump_bam_according_to_fastx_channel(args.bam, args.fastx_file, args.infix)

    pass


if __name__ == "__main__":
    cli_args = {
        "subreads_bam": "/data/ccs_data/case-study/tr-at-error-kaipu/k2_5_subreads.bam",
        "fastx_file": "/data/ccs_data/case-study/tr-at-error-kaipu/sample5.fastq",
    }

    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("--bam", type=str, required=True, dest="bam")
    parser.add_argument("--fx-file-or-bam", type=str, required=True, dest="fastx_file")
    parser.add_argument("--infix", type=str, required=True)
    # main(Namespace(**cli_args))
    main(parser.parse_args())
    pass



import pysam
import os
import sys
from tqdm import tqdm
import argparse

pre_dir = os.path.abspath(__file__).rsplit("/", maxsplit=2)[0]
print(pre_dir)
sys.path.append(pre_dir)

from fact_table_ana.utils import read_fastx_file


def compress_homopolymer(sequence):
    if not sequence:
        return "", []

    compressed = []
    counts = []

    current_char = sequence[0]
    count = 1

    for char in sequence[1:]:
        if char == current_char:
            count += 1
        else:
            compressed.append(current_char)
            counts.append(count)
            current_char = char
            count = 1

    # 添加最后一个字符和计数
    compressed.append(current_char)
    counts.append(count)

    return "".join(compressed), counts


def bam_hpc(bam_filepath: str):
    out_bam = "{}.hpc.bam".format(bam_filepath.rsplit(".", maxsplit=1)[0])

    with pysam.AlignmentFile(
        bam_filepath, mode="rb", threads=40, check_sq=False
    ) as in_bam_h:
        with pysam.AlignmentFile(
            out_bam, mode="wb", threads=40, check_sq=False, header=in_bam_h.header
        ) as out_bam_h:

            for record in tqdm(
                in_bam_h.fetch(until_eof=True), desc=f"extracting channel ref"
            ):
                record_new = pysam.AlignedSegment()
                record_new.query_name = record.query_name
                hpc_seq, counts = compress_homopolymer(record.query_sequence)
                record_new.query_sequence = hpc_seq
                record_new.set_tags(
                    [
                        ("dw", counts, "I"),
                        ("ch", int(record.get_tag("ch")), "I"),
                    ]
                )
                out_bam_h.write(record_new)


def fa_hpc(filepath: str):
    out_hpc_file = "{}.hpc.fa".format(filepath.rsplit(".", maxsplit=1)[0])
    out_hpc_cnt_file = "{}.hpc.txt".format(filepath.rsplit(".", maxsplit=1)[0])

    out_hpc_file_h = open(out_hpc_file, mode="w", encoding="utf8")
    out_hpc_cnt_file_h = open(out_hpc_cnt_file, mode="w", encoding="utf8")

    fastx_data = read_fastx_file(filepath)

    for qname, seq in fastx_data.items():
        hpc_seq, cnt = compress_homopolymer(seq[0])
        out_hpc_file_h.write(">{}\n{}\n".format(qname, hpc_seq))
        out_hpc_cnt_file_h.write(">{}\n{}\n".format(qname, ",".join(map(str, cnt))))


def main(args):
    for filepath in args.files:
        if filepath.endswith(".bam"):
            bam_hpc(filepath)
            continue
        if (
            filepath.endswith("fa")
            or filepath.endswith("fasta")
            or filepath.endswith("fna")
        ):
            fa_hpc(filepath)
            continue

        raise ValueError("invalid data format")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("files", nargs="+")

    main(args=parser.parse_args())



import mappy
import pysam
import re


def calculate_identity_from_cigar(cigar_string):
    """
    根据CIGAR字符串计算identity

    Parameters:
    cigar_string (str): CIGAR字符串, 如 "370=1I89=1D26=1I208=1D1250=1D578=1D2=2X1=1I578=1I1248=1I210=1D23=1I90=1D373="

    Returns:
    float: identity值
    """
    # 使用正则表达式解析CIGAR字符串
    pattern = r'(\d+)([=IDX])'
    matches = re.findall(pattern, cigar_string)

    # 初始化计数器
    match_count = 0      # 匹配的碱基数（=）
    mismatch_count = 0   # 错配的碱基数（X）
    total_aligned = 0    # 总比对长度（= + X + I + D）

    for length_str, operation in matches:
        length = int(length_str)

        if operation == '=':
            # 精确匹配
            match_count += length
            total_aligned += length
        elif operation == 'X':
            # 错配
            mismatch_count += length
            total_aligned += length
        elif operation == 'I':
            # 插入：参考序列没有，查询序列有
            total_aligned += length
        elif operation == 'D':
            # 删除：参考序列有，查询序列没有
            total_aligned += length

    # 计算identity
    if total_aligned == 0:
        return 0.0

    identity = match_count / total_aligned
    return identity


def read_fasta(fpath: str):
    with open(fpath, mode="r", encoding="utf8") as f:
        line = f.readline().strip()
        if line.startswith(">"):
            header = line
        seq = ""
        while True:
            new_line = f.readline().strip()
            if new_line == "":
                return header, seq
            seq += new_line


def read_fasta2(fpath):
    with pysam.FastxFile(fpath) as f:
        for record in f:
            return record.name, record.sequence


def main():
    target_seq = "GCCCCGCACTTCCACCACCAGCTCCTCCACCTTCTCTTCAGCCCTGCTAGCGCCGGGAGCCCGCCCCCGAGAGGTGGGCTGCGGGCGCTCGAGGCCCAGCCGCCGCCGCCGCCGCCGCCGCCGCCGCCTCCGCCGCCGCCGCCGCCGC"
    query_seq = "GCCCCGCACTTCCACCACCAGCTCCTCCACCTTCTCTTCAGCCCTGCTAGCGCCGGGAGCCCGCCCCCGAGAGGTGGGCTGCGGGCGCTCGAGGCCCAGCCGCCGCCGCCGCCGCCGCCGCCGCCGCCTCCGCCGCCGCCGCCGCCGC"
    aligner = mappy.Aligner(seq=target_seq, extra_flags=67108864,
                            k=9, w=7, best_n=10, n_threads=1)
    cnt = 0
    for hit in aligner.map(query_seq):
        cnt += 1
        print("ctg:{}\tr_st:{}\tr_end:{}\tq_st:{}\tq_end:{}\tstrand:{}\tref_len:{}\tquery_len:{}\tideneity:{:.4f}\tQueryCoverage:{:.4f}\tTargetCoverage:{:.4f}".format(
            hit.ctg, hit.r_st, hit.r_en, hit.q_st, hit.q_en, hit.strand, len(
                target_seq), len(query_seq),
            calculate_identity_from_cigar(hit.cigar_str),
            (hit.q_en - hit.q_st) / len(query_seq),
            (hit.r_en - hit.r_st) / len(target_seq)))

    pass


if __name__ == "__main__":
    main()




import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
import argparse
import pysam
import multiprocessing as mp


def read_pos_from_vcf_file(fname, interested_names):
    positions = []
    with open(fname, mode="r") as f:
        for line in tqdm(f, desc=f"reading {fname}"):
            line = line.strip()
            if line.startswith("#"):
                continue

            items = line.split("\t")
            if items[0].strip() in interested_names:
                pos = int(items[1])
                positions.append(pos)

    return np.array(positions)


def get_interested_channels(bam_path, length_low, length_high):
    names = set()
    tot = 0
    filtered = 0
    with pysam.AlignmentFile(filename=bam_path, check_sq=False, mode="rb", threads=mp.cpu_count()) as bam:
        for record in tqdm(bam.fetch(until_eof=True), desc=f"reading {bam_path}"):
            tot += 1
            if record.query_length >= length_low and record.query_length < length_high:
                names.add(record.query_name)
                filtered += 1
    print(f"TOT:{tot}, filtered:{filtered}")

    return names


def main():

    parser = argparse.ArgumentParser("")
    parser.add_argument("--smc-bam", required=True, dest="smc_bam")
    parser.add_argument("--vcf", required=True)
    parser.add_argument("--length-low", default=1300, dest="length_low")
    parser.add_argument("--length-high", default=1700, dest="length_high")
    args = parser.parse_args()
    
    smc_bam = args.smc_bam
    vcf_file = args.vcf
    names = get_interested_channels(smc_bam, length_low=args.length_low, length_high=args.length_high)
    positions = read_pos_from_vcf_file(vcf_file, names)

    # ================================
    #  绘制位点分布直方图
    # ================================
    plt.figure(figsize=(12, 5))
    plt.hist(positions, bins=2000)  # bins 可根据点数调整
    plt.xlabel("Genomic Position (POS)")
    plt.ylabel("Variant Count")
    plt.title("Variant Site Position Distribution")
    plt.tight_layout()
    plt.show()
    plt.savefig("vcf_position.jpg")


if __name__ == "__main__":
    main()
    pass



import semver
import subprocess
import logging
import os

logging.basicConfig(
    level=logging.INFO,
    datefmt="%Y/%m/%d %H:%M:%S",
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def install_bin(cmd):
    logging.info(f"run command: {cmd}")
    subprocess.check_call(cmd, shell=True)


def check_version(bin_name: str, expected_version):
    status, output = subprocess.getstatusoutput(f"{bin_name} -V")
    if status != 0:
        return False
    oup = output.strip()
    version_str = oup.rsplit(" ", maxsplit=1)[1]

    logging.info(f"{bin_name} Version: {version_str}")
    version = semver.Version.parse(version_str)
    if version < expected_version:
        return False

    return True


def check_and_install(bin_name: str, expected_version: semver.Version, install_cmd: str):
    if check_version(bin_name, expected_version):
        return
    install_bin(install_cmd)
    assert check_version(
        bin_name, expected_version), f"check {bin_name} version failed"


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "300"
    os.environ["POLARS_FMT_STR_LEN"] = "100"


import pathlib
import matplotlib.pyplot as plt
import seaborn as sns
import argparse
import polars as pl
import os
import sys

cur_dir = os.path.abspath(__file__).rsplit("/", maxsplit=1)[0]
sys.path.insert(0, cur_dir)
import polars_init  # noqa: E402
import utils  # noqa: E402


def main(args):
    polars_init.polars_env_init()

    # plt.grid(True, linestyle=":", linewidth=0.5, color="gray")
    fact_table_path = pathlib.Path(args.fact_table)
    df = pl.read_csv(args.fact_table, separator="\t")
    df = df.with_columns(
        [
            (pl.col("eq") / (pl.col("eq") + pl.col("diff") + pl.col("ins") + pl.col("del"))).alias(
                "emp_rq"
            )
        ]
    ).with_columns([utils.q2phreq_expr("emp_rq", "emp_phreq")])
    figure = plt.figure(figsize=(10, 10))
    axs = figure.add_subplot(1, 1, 1)
    plt.sca(axs)
    plt.grid(True, linestyle=":", linewidth=0.5, color="gray")

    sns.scatterplot(df.to_pandas(), x="baseq", y="emp_phreq", ax=axs)
    axs.set_xticks(list(range(0, 60, 2)))
    axs.set_yticks(list(range(0, 60, 2)))
    axs.set_xlabel("PredictedBaseQ", fontdict={"size": 16})
    axs.set_ylabel("EmpericalBaseQ", fontdict={"size": 16})
    perfect_line = pl.DataFrame(
        {
            "x": list(range(0, 60)),
            "y": list(range(0, 60)),
        }
    )

    sns.lineplot(
        perfect_line.to_pandas(), x="x", y="y", ax=axs, color="blue", linestyle="--"
    )

    print(df.head(60))

    summary = df.select([
        (pl.col("depth").filter(pl.col("baseq") >= 20).sum() /
         pl.col("depth").sum()).alias("baseq20Ratio"),
        (pl.col("depth").filter(pl.col("baseq") >= 30).sum() /
         pl.col("depth").sum()).alias("baseq30Ratio"),
        (pl.col("depth").filter(pl.col("baseq") >= 35).sum() /
         pl.col("depth").sum()).alias("baseq35Ratio"),
    ])
    print("Base Q Summary:\n", summary.transpose(
        include_header=True, header_name="name", column_names=["value"]
    ))

    metric = (df.filter(pl.col("depth") > 10000)
              .select([(pl.col("baseq") - pl.col("emp_phreq")).pow(2.0).alias("SquareError"),
                       (pl.col("baseq") - pl.col("emp_phreq")
                        ).abs().alias("AbsError"),
                       ((pl.col("emp_phreq") - pl.col("baseq")).abs() /
                        pl.col("emp_phreq")).alias("ape"),
                       (2 * (pl.col("baseq")-pl.col("emp_phreq")).abs() /
                        (pl.col("baseq")+pl.col("emp_phreq"))).alias("sape")
                       ])
              .select([
                  pl.col("SquareError").mean().alias("MSE"),
                  pl.col("SquareError").mean().sqrt().alias("RMSE"),
                  pl.col("AbsError").mean().alias("MAE"),
                  pl.col("AbsError").median().alias("MedAE"),
                  pl.col("ape").mean().alias("MAPE"),
                  pl.col("sape").mean().alias("sMAPE"),

              ])
              )
    
    metric = metric.transpose(
        include_header=True, header_name="name", column_names=["value"]
    )
    print("Base Q Metric: \n", metric)
    

    baseq2emp_baseq_fpath = f"{args.o_prefix}.baseq2empq.png"
    figure.savefig(fname=baseq2emp_baseq_fpath)
    print(f"check image {baseq2emp_baseq_fpath}")
    
    baseq_cnt = (df.filter(pl.col("depth") > 10000)
        .select([pl.col("baseq"), pl.col("depth")]))
    baseq_cnt = baseq_cnt.to_pandas()
    
    
    
    # TODO baseq distribution
    figure = plt.figure(figsize=(10, 10))
    axs = figure.add_subplot(1, 1, 1)
    plt.sca(axs)
    plt.grid(True, linestyle=":", linewidth=0.5, color="gray")
    sns.barplot(baseq_cnt, x="baseq", y="depth", ax=axs, order=list(range(0, 50)),)

    axs.set_xticks(list(range(0, 50, 2)))
    axs.set_xlabel("PredictedBaseQ", fontdict={"size": 16})
    axs.set_ylabel("Count", fontdict={"size": 16})
    
    baseq_dist_fpath = f"{args.o_prefix}.baseq-dist.png"
    figure.savefig(fname=baseq_dist_fpath)
    print(f"check image {baseq_dist_fpath}")
    
    
    
    
    


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="",
        usage="""
    gsmm2 align -q $query_file -t $ref_file -p outputbam_prefix --noMar
    gsetl --outdir $outdir aligned-bam --bam $aligned_bam --ref-file $ref_file
    python preq-baseq-and-emp-q.py $outdir/fact_baseq_stat.csv
""",
    )
    parser.add_argument("fact_table", metavar="fact_baseq_stat")
    parser.add_argument("--o-path", metavar="o-path",
                        default=None, dest="o_path")

    main(parser.parse_args())



import pysam
from tqdm import tqdm
from typing import Set, Mapping, List


def extract_interested_channels(bam_filepath: str, min_length: int, max_length=None):
    channels = set()
    with pysam.AlignmentFile(
        filename=bam_filepath, mode="rb", check_sq=False, threads=40
    ) as bam_file:
        for read in tqdm(
            bam_file.fetch(until_eof=True), desc=f"reading {bam_filepath}"
        ):
            if read.query_name.endswith("sbr"):
                continue

            if read.get_tag("np") < 3:
                continue
            if read.query_length < min_length:
                continue
            if max_length is not None and read.query_length > max_length:
                continue
            channels.add(read.get_tag("ch"))
    return channels


def expand_channels(channels: Set[int], expected_num: int) -> Mapping[int, List[int]]:
    channels = sorted(list(channels))
    num_ch = len(channels)
    result = {}
    for new_ch in range(expected_num):
        old_ch_idx = new_ch % num_ch
        old_ch = channels[old_ch_idx]
        result.setdefault(old_ch, []).append(new_ch)
    return result


def modify_read_info(read_old: pysam.AlignedSegment, old_ch: int, new_ch: int):
    idx = read_old.query_name.rsplit("/", maxsplit=1)[1]
    read_old.query_name = f"read_{new_ch}/{new_ch}/subread/{idx}"
    read_old.set_tag("ch", new_ch, value_type="i")
    return read_old


def dumping_new_subreads(sbr_bam_filepath: str, channels: Mapping[int, List[int]], min_length, max_length=None):
    o_bam_filename = "{}.{}.{}-{}.bam".format(
        sbr_bam_filepath.rsplit(".", maxsplit=1)[
            0], "expanded-2", min_length, max_length
    )
    o_bam_sorted_filename = "{}.{}.{}-{}.sorted.bam".format(
        sbr_bam_filepath.rsplit(".", maxsplit=1)[
            0], "expanded-2", min_length, max_length
    )
    with pysam.AlignmentFile(
        sbr_bam_filepath, mode="rb", threads=40, check_sq=False
    ) as in_bam:
        with pysam.AlignmentFile(
            o_bam_filename, mode="wb", threads=40, check_sq=False, header=in_bam.header
        ) as out_bam:

            for record in tqdm(
                in_bam.fetch(until_eof=True), desc=f"dumping {o_bam_filename}"
            ):
                ch = int(record.get_tag("ch"))
                if ch in channels:
                    old_ch = ch
                    for new_ch in channels[ch]:
                        record = modify_read_info(
                            read_old=record, old_ch=old_ch, new_ch=new_ch
                        )
                        out_bam.write(record)
                        old_ch = new_ch
    print("sorting....")
    pysam.sort("-n", "-t", "ch", "-@", "100", "-o",
               o_bam_sorted_filename, o_bam_filename)


def main_cli():

    sbr_bam_filepath = "/data/ccs_data/speed-test/benchmark-data/20250724_240601Y0002_Run0005_adapter.5h5k.bam"
    smc_bam_filepath = "/data/ccs_data/speed-test/benchmark-data/20250724_240601Y0002_Run0005_adapter.smc_all_reads.bam"
    min_length = 3000
    max_length = 7000
    channels = extract_interested_channels(
        smc_bam_filepath, min_length=min_length, max_length=max_length)
    print(f"OriChannelNum:{len(channels)}")

    channels = expand_channels(channels=channels, expected_num=500000)
    dumping_new_subreads(sbr_bam_filepath=sbr_bam_filepath,
                         channels=channels, min_length=min_length, max_length=max_length)


if __name__ == "__main__":
    main_cli()



import polars as pl
import argparse

import os
import sys

cur_dir = os.path.abspath(__file__).rsplit("/", maxsplit=1)[0]
sys.path.insert(0, cur_dir)

import polars_init


def error_channel_analysis(
    fact_aligned_bam_bam_basic: str, fact_error_query_locus_info: str
):

    basic = pl.read_csv(fact_aligned_bam_bam_basic, separator="\t")
    query_error_locus = pl.read_csv(fact_error_query_locus_info, separator="\t")

    basic = basic.filter((pl.col("np") == 7).and_(pl.col("iy") < 0.999)).select(
        [
            pl.col("qname"),
            pl.col("np"),
            pl.col("rq"),
            pl.col("iy"),
            pl.col("qlen"),
            pl.col("fwd"),
        ]
    )

    df = (
        basic.join(query_error_locus, on="qname", how="inner")
        .with_columns(
            [
                pl.when(pl.col("fwd"))
                .then(pl.col("qstart"))
                .otherwise(pl.col("qlen") - pl.col("qend"))
                .alias("qstart"),
                pl.when(pl.col("fwd"))
                .then(pl.col("qend"))
                .otherwise(pl.col("qlen") - pl.col("qstart"))
                .alias("qend"),
            ]
        )
        .sort(by=["qname", "rstart"], descending=[True, False])
    )
    print(df.head(100))


def main(args):
    error_channel_analysis(args.bam_basic, args.error_query_locus)
    pass


if __name__ == "__main__":
    polars_init.polars_env_init()

    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("bam_basic")
    parser.add_argument("error_query_locus")
    main(parser.parse_args())



import pysam
import sys
from tqdm import tqdm

def find_high_qual_reads(bam_file, min_base_qual=45, max_reads=100):
    """
    查找BAM文件中存在base quality > min_base_qual的reads
    
    参数:
    bam_file: BAM文件路径
    min_base_qual: 最小base quality阈值
    max_reads: 最大输出reads数量
    """
    
    # 打开BAM文件
    bam = pysam.AlignmentFile(bam_file, "rb", check_sq=False, threads=40)
    
    count = 0
    found_reads = 0
    
    print(f"搜索BAM文件中存在base quality > {min_base_qual}的reads...")
    print("=" * 80)
    
    try:
        for read in tqdm(bam.fetch(until_eof=True), desc=f"reading {bam_file}"):
            
            if read.get_tag("np") < 3:
                continue
            
            count += 1
            
            # 检查read是否有quality信息
            if read.query_qualities is None:
                continue
            
            # 检查是否有任何一个base的quality > 阈值
            has_high_qual = any(qual > min_base_qual for qual in read.query_qualities)
            
            if has_high_qual:
                found_reads += 1
                print(f"\nRead #{found_reads}:")
                print(f"Read Name: {read.query_name}")
                print(f"Flags: {read.flag} (0x{read.flag:04x})")
                
                # 打印quality值统计
                quals = read.query_qualities
                max_qual = max(quals) if quals else 0
                min_qual = min(quals) if quals else 0
                avg_qual = sum(quals) / len(quals) if quals else 0
                
                print(f"Quality Stats - Max: {max_qual}, Min: {min_qual}, Avg: {avg_qual:.2f}")
                
                # 打印前20个base的quality值（如果read很长）
                print("First 20 base qualities:")
                qual_str = " ".join(f"{q:2d}" for q in quals[:20])
                print(qual_str, end=" ")
                
                if len(quals) > 20:
                    print(f"... and {len(quals) - 20} more bases")
                    
                print(read)
                
                print("-" * 60)
                
                # 如果找到的reads数量达到限制，停止搜索
                if found_reads >= max_reads:
                    print(f"\n已达到最大显示数量 ({max_reads})，停止搜索。")
                    break
            
            # 每处理10000个reads打印进度
            if count % 10000 == 0:
                print(f"已处理 {count} 个reads, 找到 {found_reads} 个符合条件的reads...")
    
    except KeyboardInterrupt:
        print("\n用户中断搜索")
    
    finally:
        bam.close()
    
    print(f"\n统计信息:")
    print(f"总共处理的reads: {count}")
    print(f"找到的符合条件的reads: {found_reads}")
    print(f"比例: {found_reads/count*100:.2f}%")

def detailed_quality_analysis(bam_file, min_base_qual=45, sample_size=50):
    """
    对符合条件的reads进行更详细的质量分析
    """
    bam = pysam.AlignmentFile(bam_file, "rb", check_sq=False, threads=40)
    
    high_qual_reads = []
    count = 0
    
    print(f"\n进行详细质量分析 (采样 {sample_size} 个reads)...")
    print("=" * 80)
    
    for read in tqdm(bam.fetch(until_eof=True), desc=f"reading {bam_file}"):
        
        if read.get_tag("np") < 3:
                continue
        
        if read.query_qualities is None:
            continue
        
        if any(qual > min_base_qual for qual in read.query_qualities):
            high_qual_reads.append(read)
            count += 1
            
            if count >= sample_size:
                break
    
    bam.close()
    
    if not high_qual_reads:
        print("未找到符合条件的reads")
        return
    
    # 分析质量分布
    all_high_quals = []
    for read in high_qual_reads:
        high_quals = [q for q in read.query_qualities if q > min_base_qual]
        all_high_quals.extend(high_quals)
    
    if all_high_quals:
        print(f"高质量base (> {min_base_qual}) 的统计:")
        print(f"  总数: {len(all_high_quals)}")
        print(f"  平均质量: {sum(all_high_quals)/len(all_high_quals):.2f}")
        print(f"  最大质量: {max(all_high_quals)}")
        print(f"  最小质量: {min(all_high_quals)}")
        
        # 质量值分布
        qual_bins = {}
        for qual in all_high_quals:
            bin_key = f"{qual}"
            qual_bins[bin_key] = qual_bins.get(bin_key, 0) + 1
        
        print(f"\n质量值分布:")
        for qual in sorted(qual_bins.keys(), key=int, reverse=True):
            print(f"  Q{qual}: {qual_bins[qual]}个base")

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("用法: python script.py <bam_file> [min_base_qual] [max_reads]")
        print("示例: python script.py sample.bam 45 50")
        sys.exit(1)
    
    bam_file = sys.argv[1]
    min_base_qual = int(sys.argv[2]) if len(sys.argv) > 2 else 45
    max_reads = int(sys.argv[3]) if len(sys.argv) > 3 else 100
    
    # 检查BAM文件是否存在
    try:
        # 首先检查BAM索引文件
        pysam.index(bam_file)
    except:
        print(f"注意: BAM文件可能需要索引, 请确保 {bam_file}.bai 文件存在")
    
    # 查找高质量reads
    find_high_qual_reads(bam_file, min_base_qual, max_reads)
    
    # 进行详细分析
    detailed_quality_analysis(bam_file, min_base_qual)


import subprocess
import pathlib
import os
import logging
import polars as pl
import shutil
import argparse
from multiprocessing import cpu_count
import os
import semver
import threading
import matplotlib.pyplot as plt
import seaborn as sns
from glob import glob

# deprecated ...
logging.basicConfig(
    level=logging.INFO,
    datefmt="%Y/%m/%d %H:%M:%S",
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "300"
    os.environ["POLARS_FMT_STR_LEN"] = "100"


def mm2_version_check():
    oup = subprocess.getoutput("gsmm2-aligned-metric -V")
    oup = oup.strip()
    version_str = oup.rsplit(" ", maxsplit=1)[1]

    logging.info(f"gsmm2-aligned-metric Version: {version_str}")
    mm2_version = semver.Version.parse(version_str)
    expected_version = "0.21.0"
    assert mm2_version >= semver.Version.parse(
        expected_version
    ), f"current mm2 version:{mm2_version} < {expected_version}, try 'cargo uninstall mm2; cargo install mm2@={expected_version}' "


def gsetl_version_check():
    oup = subprocess.getoutput("gsetl -V")
    oup = oup.strip()
    version_str = oup.rsplit(" ", maxsplit=1)[1]

    logging.info(f"gsetl Version: {version_str}")
    gsetl_version = semver.Version.parse(version_str)
    expected_version = "0.7.1"
    assert gsetl_version >= semver.Version.parse(
        expected_version
    ), f"current gsetl version:{gsetl_version} < {expected_version}, try 'cargo uninstall gsetl; cargo install gsetl@={expected_version}' "


def extract_filename(filepath: str) -> str:
    p = pathlib.Path(filepath)
    return p.stem


def generate_aligned_metric_fact_file(
    bam_file: str,
    ref_fasta: str,
    out_filename: str,
    short_aln=False,
    force: bool = False,
    threads=None,
    no_supp=False,
    no_mar=False,
) -> str:

    if no_supp and no_mar:
        raise ValueError("no_supp, no_mar can't be all true")

    if force and os.path.exists(out_filename):
        os.remove(out_filename)

    if os.path.exists(out_filename):
        logging.warning("use the existing metric file : %s", out_filename)
        return out_filename

    threads = cpu_count() if threads is None else threads
    cmd = f"""gsmm2-aligned-metric --threads {threads} \
            -q {bam_file} \
            -t {ref_fasta} \
            --out {out_filename} \
            --kmer 11 \
            --wins 1 """
    if short_aln:
        cmd += "  --short-aln"

    logging.info("cmd: %s", cmd)
    subprocess.check_call(cmd, shell=True)

    return out_filename


def analysis_aligned(df: pl.DataFrame) -> pl.DataFrame:
    df = (
        df.select(
            [
                pl.when(pl.col("rname").eq(pl.lit("")).or_(
                    pl.col("rname").is_null()))
                .then(pl.lit("notAligned"))
                .otherwise(pl.lit("aligned"))
                .alias("name")
            ]
        )
        .group_by(["name"])
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [(pl.col("cnt") / pl.col("cnt").sum().over(pl.lit(1))).alias("ratio")]
        )
    )

    metric_cnt = df.select(
        [pl.col("name"), pl.col("cnt").cast(pl.Float64).alias("value")]
    ).sort(by=["name"])

    metric_ratio = df.select(
        [pl.format("{}Ratio", pl.col("name")), pl.col("ratio").alias("value")]
    ).sort(by=["name"])

    return pl.concat([metric_cnt, metric_ratio])


def analysis_segs(df: pl.DataFrame) -> pl.DataFrame:
    return (
        df.with_columns(
            [
                pl.when(pl.col("segs") >= 20)
                .then(pl.lit(20))
                .otherwise(pl.col("segs"))
                .alias("segs"),
                pl.when(pl.col("segs") >= 20)
                .then(pl.lit("≥20"))
                .otherwise(pl.format("={}", pl.col("segs")))
                .alias("tag"),
            ]
        )
        .group_by(["segs", "tag"])
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [
                (pl.col("cnt") / pl.col("cnt").sum().over(pl.lit("1"))).alias("ratio"),
                pl.col("cnt").sum().over(pl.lit("1")).alias("tot_cnt"),
            ]
        )
        .sort("segs")
        .select(
            [
                pl.format("[SegsRatio]segs{}", pl.col("tag")).alias("name"),
                pl.col("ratio").alias("value"),
            ]
        )
    )


def analysis_segs2(
    df: pl.DataFrame,
    ovlp_low_threshold: float = 0.02,
    ovlp_high_threshold: float = 0.9,
) -> pl.DataFrame:
    return (
        df.filter(pl.col("segs") == pl.lit(2))
        .with_columns(
            [
                (pl.col("qOvlpRatio") < ovlp_low_threshold).alias("nonOvlpQuery"),
            ]
        )
        .with_columns(
            [
                (
                    pl.col("nonOvlpQuery").and_(
                        pl.col("rOvlpRatio") > ovlp_high_threshold
                    )
                ).alias("svCandidate"),
                (
                    pl.col("nonOvlpQuery").and_(
                        pl.col("rOvlpRatio") < ovlp_low_threshold
                    )
                ).alias("noCutCandidate"),
            ]
        )
        .with_columns(
            [pl.col("oriQGaps").str.split(",").list.get(
                1).cast(pl.Int32).alias("gap")]
        )
        .with_columns([pl.col("gap").lt(pl.lit(20)).alias("gap<20")])
        .with_columns(
            [
                pl.col("svCandidate").and_(pl.col("gap<20")).alias("sv"),
                pl.col("noCutCandidate").and_(pl.col("gap<20")).alias("noCut"),
            ]
        )
        .with_columns(
            [
                pl.when(pl.col("sv"))
                .then(pl.lit("sv"))
                .when(pl.col("noCut"))
                .then(pl.lit("noCut"))
                .when(pl.col("svCandidate"))
                .then(pl.lit("svCandidate"))
                .when(pl.col("noCutCandidate"))
                .then(pl.lit("noCutCandidate"))
                .otherwise(pl.lit("badCase"))
                .alias("tag")
            ]
        )
        .with_columns(
            [
                pl.when(
                    pl.col("tag").eq(pl.lit("badCase")).and_(
                        pl.col("identity") < 0.85)
                )
                .then(pl.lit("badCase-lowIdentity"))
                .otherwise(pl.col("tag"))
                .alias("tag")
            ]
        )
        .group_by(["tag"])
        .agg([pl.len().alias("cnt")])
        .select(
            [
                pl.col("tag"),
                pl.col("cnt"),
                pl.col("cnt").sum().over(pl.lit("1")).alias("tot_cnt"),
            ]
        )
        .with_columns([(pl.col("cnt") / pl.col("tot_cnt")).alias("ratio")])
        .sort(["tag"])
        .select(
            [
                pl.format("[segs=2]{}", pl.col("tag")).alias("name"),
                pl.col("ratio").alias("value"),
            ]
        )
    )


def analysis_gaps(df: pl.DataFrame) -> pl.DataFrame:

    segs_ratio = (
        df.select(
            [
                pl.when(pl.col("segs") > 1)
                .then(pl.lit("[segs>1]"))
                .otherwise(pl.lit("[segs=1]"))
                .alias("name")
            ]
        )
        .group_by("name")
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [(pl.col("cnt") / pl.col("cnt").sum().over(pl.lit(1))).alias("ratio")]
        )
    )

    metric_ratio = pl.concat(
        [
            segs_ratio.filter(pl.col("name").eq(pl.lit("[segs>1]"))).select(
                [
                    pl.format("{}Cnt", pl.col("name")),
                    pl.col("cnt").cast(pl.Float64).alias("value"),
                ]
            ),
            segs_ratio.filter(pl.col("name").eq(pl.lit("[segs>1]"))).select(
                [
                    pl.format("{}Ratio", pl.col("name")),
                    pl.col("ratio").cast(pl.Float64).alias("value"),
                ]
            ),
        ]
    )

    top20_gap = (
        df.filter(pl.col("segs") > 1)
        .select(
            [
                pl.col("oriQGaps")
                .str.split(",")
                .list.slice(1, pl.col("segs") - 1)
                .explode()
                .alias("gap")
            ]
        )
        .select([pl.col("gap").cast(pl.Int32)])
        .group_by("gap")
        .agg([pl.len().alias("cnt")])
        .with_columns(
            [(pl.col("cnt") / pl.col("cnt").sum().over(pl.lit(1))).alias("ratio")]
        )
        .sort(["ratio"], descending=[True])
        .head(20)
    )

    top20_gap_metric = top20_gap.select(
        [
            pl.format("[segs>1]gap={}", pl.col("gap")).alias("name"),
            pl.col("ratio").alias("value"),
        ]
    )

    top20_tot = top20_gap.select([pl.col("ratio").sum().alias("value")]).select(
        [pl.lit("[segs>1]gapTop20Ratio").alias("name"), pl.col("value")]
    )

    # .filter(pl.col("ratio") > 0.01)
    # .sort("gap")
    # .select(
    #     [
    #         pl.format("[segs>1]gap={}", pl.col("gap")).alias("name"),
    #         pl.col("ratio").alias("value"),
    #     ]
    # )
    return pl.concat([metric_ratio, top20_tot, top20_gap_metric])


def analisys_long_indel(df: pl.DataFrame) -> pl.DataFrame:
    metric = df.select(
        [
            pl.len().alias("cnt"),
            pl.col("longIndel")
            .filter(pl.col("longIndel").is_not_null())
            .len()
            .alias("longIndelCnt"),
        ]
    ).select(
        [
            pl.lit("longIndelRatio").alias("name"),
            (pl.col("longIndelCnt") / pl.col("cnt")).alias("value"),
        ]
    )
    return metric


def stats(metric_filename: str, filename: str):
    df = pl.read_csv(
        metric_filename, separator="\t", infer_schema_length=3000, schema_overrides={"longIndel": pl.String}
    )
    metric_aligned_not_aligned = analysis_aligned(df=df)
    df = df.filter(pl.col("rname") != "")
    metric_long_indel = analisys_long_indel(df=df)
    metric_segs = analysis_segs(df)
    metric_segs2 = analysis_segs2(df)
    metric_gaps = analysis_gaps(df=df)

    df = df.with_columns(
        [
            ((pl.col("qOvlpRatio") < 0.01).and_(pl.col("rOvlpRatio") < 0.01))
            .or_(
                (pl.col("qOvlpRatio") < 0.01)
                .and_(pl.col("rOvlpRatio") > 0.90)
                .and_(pl.col("oriQGaps").str.split(",").list.get(1).cast(pl.Int32) < 20)
            )
            .alias("valid")
        ]
    )

    df = df.with_columns(
        [
            pl.when(pl.col("valid"))
            .then(pl.col("covlen"))
            .otherwise(pl.col("primaryCovlen"))
            .alias("miscCovlen")
        ]
    )

    df = df.with_columns(row_align_span())
    aggr_metrics = df.select(aggr_expressions())
    aggr_metrics = aggr_metrics.transpose(
        include_header=True, header_name="name", column_names=["value"]
    )

    aggr_metrics = pl.concat(
        [
            metric_aligned_not_aligned,
            metric_long_indel,
            aggr_metrics,
            metric_segs,
            metric_segs2,
            metric_gaps,
        ]
    )

    print(aggr_metrics)

    aggr_metrics.write_csv(filename, include_header=True, separator="\t")
    identity_min = 0.6
    identity_max = 1.0
    if df.shape[0] > 10000:
        sampled_identity = df.select([pl.col("identity").clip(
            lower_bound=identity_min, upper_bound=identity_max)]).sample(n=10000, seed=2025).to_pandas()
    else:
        sampled_identity = df.to_pandas()

    identity_hist_filename = "{}.idenity_hist.png".format(
        filename.rsplit(".", maxsplit=1)[0])

    plot_histgram(sampled_identity["identity"], fname=identity_hist_filename, xlabel="Identity",
                  ylabel="Count", title="SampledIdentityHist", xlim=(identity_min, identity_max))


def aggr_expressions():

    exprs = [
        pl.col("covlen").sum().cast(pl.Float64).alias("alignedBases"),
        (pl.col("primaryCovlen").sum() / pl.col("qlen").sum()).alias("queryCoverage"),
        (pl.col("miscCovlen").sum() / pl.col("qlen").sum()).alias("queryCoverage2"),
        (pl.col("covlen").sum() / pl.col("qlen").sum()).alias("queryCoverage3"),
        pl.quantile("queryCoverage", 0.25).alias("queryCoverage-p25"),
        pl.col("queryCoverage").median().alias("queryCoverage-p50"),
        pl.quantile("queryCoverage", 0.75).alias("queryCoverage-p75"),
        (pl.col("match").sum() / pl.col("alignSpan").sum()).alias("identity"),
        pl.quantile("identity", 0.25).alias("identity-p25"),
        pl.col("identity").median().alias("identity-p50"),
        pl.quantile("identity", 0.75).alias("identity-p75"),
        (pl.col("misMatch").sum() / pl.col("alignSpan").sum()).alias("mmRate"),
        (pl.col("ins").sum() / pl.col("alignSpan").sum()).alias("NHInsRate"),
        (pl.col("homoIns").sum() / pl.col("alignSpan").sum()).alias("HomoInsRate"),
        ((pl.col("ins").sum() + pl.col("homoIns").sum()) /
         pl.col("alignSpan").sum()).alias("insRate"),

        (pl.col("del").sum() / pl.col("alignSpan").sum()).alias("NHDelRate"),
        (pl.col("homoDel").sum() / pl.col("alignSpan").sum()).alias("HomoDelRate"),
        ((pl.col("homoDel").sum() + pl.col("del").sum()) /
         pl.col("alignSpan").sum()).alias("delRate"),
    ]

    for base in "ACGT":
        exprs.extend(
            [
                (
                    pl.col(f"match-{base}").sum() /
                    pl.col(f"alignSpan-{base}").sum()
                ).alias(f"identity-{base}"),
                (
                    pl.col(f"misMatch-{base}").sum() /
                    pl.col(f"alignSpan-{base}").sum()
                ).alias(f"mmRate-{base}"),
                (pl.col(f"ins-{base}").sum() / pl.col(f"alignSpan-{base}").sum()).alias(
                    f"NHInsRate-{base}"
                ),
                (
                    pl.col(f"homoIns-{base}").sum() /
                    pl.col(f"alignSpan-{base}").sum()
                ).alias(f"HomoInsRate-{base}"),

                (
                    (pl.col(f"homoIns-{base}").sum() + pl.col(f"ins-{base}").sum()
                     ) / pl.col(f"alignSpan-{base}").sum()
                ).alias(f"insRate-{base}"),

                (pl.col(f"del-{base}").sum() / pl.col(f"alignSpan-{base}").sum()).alias(
                    f"NHDelRate-{base}"
                ),
                (
                    pl.col(f"homoDel-{base}").sum() /
                    pl.col(f"alignSpan-{base}").sum()
                ).alias(f"HomoDelRate-{base}"),

                (
                    (pl.col(f"homoDel-{base}").sum() + pl.col(f"del-{base}").sum()
                     ) / pl.col(f"alignSpan-{base}").sum()
                ).alias(f"delRate-{base}"),
            ]
        )

    return exprs


def row_align_span():
    exprs = [
        (
            pl.col("match")
            + pl.col("misMatch")
            + pl.col("ins")
            + pl.col("homoIns")
            + pl.col("del")
            + pl.col("homoDel")
        ).alias("alignSpan")
    ]

    for base in "ACGT":
        exprs.append(
            (
                pl.col(f"match-{base}")
                + pl.col(f"misMatch-{base}")
                + pl.col(f"ins-{base}")
                + pl.col(f"homoIns-{base}")
                + pl.col(f"del-{base}")
                + pl.col(f"homoDel-{base}")
            ).alias(f"alignSpan-{base}")
        )
    return exprs


def aligned_metric_analysis(fact_metric_filename: str, aggr_metric_filename: str, force: bool):
    if force and os.path.exists(aggr_metric_filename):
        os.remove(aggr_metric_filename)

    if os.path.exists(aggr_metric_filename):
        logging.warning(f"{aggr_metric_filename} will be override")

    # if not os.path.exists(aggr_metric_filename):
    stats(fact_metric_filename, filename=aggr_metric_filename)


def generate_non_aligned_metric_fact_file(bam_file: str, out_filepath: str, out_dir: str, force: bool):
    if not force and os.path.exists(out_filepath):
        logging.info(f"{out_filepath} exists, use the existed file")
        return

    cmd = f"gsetl --outdir {out_dir} non-aligned-bam --bam {bam_file} -o {out_filepath}"
    logging.info("cmd: %s", cmd)
    subprocess.check_call(cmd, shell=True)


def non_aligned_metric_analysis(fact_metric_filename: str, aggr_metric_filename: str, force: bool, out_dir: str, stem: str):
    if os.path.exists(aggr_metric_filename) and not force:
        logging.warning(f"{aggr_metric_filename} will be override")

    df = pl.read_csv(
        fact_metric_filename, separator="\t",
        infer_schema_length=3000,
        schema_overrides={
            "oe": pl.Float32
        })

    df = df.with_columns([
        (pl.col("dw_sum") * pl.lit(2)).alias("dw_sum"),
        (pl.col("ar_sum") * pl.lit(2)).alias("ar_sum"),
    ])

    whole_aggr = df.select([
        (pl.col("dw_sum").sum() / pl.col("base_cnt").sum()).alias("dw-mean"),
        (pl.col("ar_sum").sum() / pl.col("base_cnt").sum()).alias("ar-mean"),
        pl.col("cq").mean().alias("cq-mean"),
        ((pl.col("base_cnt") * pl.col("cr_mean")).sum() /
         pl.col("base_cnt").sum()).alias("cr-mean"),
        pl.col("oe").median().alias("oe-median"),
        (pl.col("base_cnt").sum() / ((pl.col("dw_sum").sum() +
         pl.col("ar_sum").sum()) * pl.lit(0.001))).alias("speed")
    ])

    base_level_aggr = df.group_by(["base"]).agg([
        (pl.col("dw_sum").sum() / pl.col("base_cnt").sum()).alias("dw-mean"),
        (pl.col("ar_sum").sum() / pl.col("base_cnt").sum()).alias("ar-mean")
    ])

    whole_aggr = whole_aggr.transpose(
        include_header=True, header_name="name", column_names=["value"]
    )

    metrics = [whole_aggr]

    for base in ["A", "C", "G", "T"]:
        tmp = base_level_aggr.filter([pl.col("base") == pl.lit(base)])\
            .select([pl.col("dw-mean").alias(f"dw-mean-{base}"), pl.col("ar-mean").alias(f"ar-mean-{base}")])
        metrics.append(tmp.transpose(
            include_header=True, header_name="name", column_names=["value"]))

    metrics = pl.concat(metrics)

    print(metrics)

    metrics.write_csv(aggr_metric_filename,
                      include_header=True, separator="\t")

    # TODO draw plot
    read_lengths = df.group_by(["qname"])\
        .agg([pl.col("base_cnt").sum()])\
        .select([pl.col("base_cnt")])
    if read_lengths.shape[0] > 10000:
        read_lengths = read_lengths.sample(
            n=10000, seed=2025).to_pandas()["base_cnt"]
    else:
        read_lengths = read_lengths.to_pandas()["base_cnt"]

    read_length_hist_fname = f"{out_dir}/{stem}.readlength_hist.png"
    plot_histgram(read_lengths, read_length_hist_fname,
                  xlabel="ReadLength", ylabel="Count", title="SampledReadLengthHist")

    dw_hist_fname = f"{out_dir}/{stem}.dw_hist.png"
    dw_min = 0
    dw_max = 200
    dw = df.group_by(["qname"])\
        .agg([pl.col("base_cnt").sum(), pl.col("dw_sum").sum()])\
        .select([(pl.col("dw_sum") / pl.col("base_cnt")).alias("dw")])\
        .select([pl.col("dw").clip(lower_bound=dw_min, upper_bound=dw_max)])
    if dw.shape[0] > 10000:
        dw = dw.sample(n=10000, seed=2025).to_pandas()["dw"]
    else:
        dw = dw.to_pandas()["dw"]
    plot_histgram(dw, dw_hist_fname, xlabel="DwellTime", ylabel="Count",
                  title="SampledDwellTimeHist", xlim=(dw_min, dw_max))

    ar_min = 0
    ar_max = 1000
    ar = df.group_by(["qname"])\
        .agg([pl.col("base_cnt").sum(), pl.col("ar_sum").sum()])\
        .select([(pl.col("ar_sum") / pl.col("base_cnt")).alias("ar")])\
        .select([pl.col("ar").clip(lower_bound=ar_min, upper_bound=ar_max)])
    if ar.shape[0] > 10000:
        ar = ar.sample(n=10000, seed=2025).to_pandas()["ar"]
    else:
        ar = ar.to_pandas()["ar"]

    ar_hist_fname = f"{out_dir}/{stem}.ar_hist.png"
    plot_histgram(ar, ar_hist_fname, xlabel="ArrivalTime", ylabel="Count",
                  title="SampledArrivalTimeHist", xlim=(ar_min, ar_max))

    pass


def plot_histgram(data, fname, xlabel, ylabel, title, xlim=None, bins=100):
    plt.figure(figsize=(8, 6))

    # 绘制直方图
    sns.histplot(data, bins=bins, kde=True, color='skyblue', binrange=xlim)

    # 添加标题与标签
    plt.title(title, fontsize=16)
    plt.xlabel(xlabel, fontsize=14)
    if xlim is not None:
        plt.xlim(xlim)
    plt.ylabel(ylabel, fontsize=14)

    # 保存图片到文件
    plt.savefig(fname, dpi=300, bbox_inches='tight')


def main(
    bam_file: str,
    ref_fa: str,
    short_aln=False,
    threads=None,
    force=False,
    outdir=None,
    copy_bam_file=False,
    enable_basic=True,
    enable_align=True
) -> str:
    
    """
        step1: generate detailed metric info
        step2: compute the aggr metric. the result aggr_metric.csv is a '\t' seperated csv file. the header is name\tvalue
            here is a demo.
            ---------aggr_metric.csv
            name    value
            queryCoverage   0.937
            ----------

    requirements:
        mm2: cargo install mm2 (version >= 0.19.0)

    Args:
        bam_file (str): bam file. only support adapter.bam
        ref_fa (str): ref genome fa file nam
        threads (int|None): threads for generating detailed metric file
        force (boolean): if force==False, use the existing metric file if exists
        outdir: if None, ${bam_filedir}/${bam_file_stem}-metric as outdir
        copy_bam_file: copy bam file to outdir. Set this parameter to true when the file is on the NAS.

    Return:
        (aggr_metric_filename, fact_metric_filename) (str, str)
    """
    polars_env_init()
    

    mm2_version_check()
    gsetl_version_check()

    if copy_bam_file:
        assert outdir is not None, "must provide outdir when copy_bam_file=True"
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        new_bam_file = os.path.join(outdir, os.path.basename(bam_file))
        if os.path.exists(new_bam_file):
            raise ValueError(f"{new_bam_file} already exists")
        shutil.copy(bam_file, new_bam_file)
        bam_file = new_bam_file

    bam_filedir = os.path.dirname(bam_file)
    stem = extract_filename(bam_file)
    if outdir is None:
        outdir = os.path.join(bam_filedir, f"{stem}-metric")

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    fact_metric_filename = f"{outdir}/{stem}.gsmm2_aligned_metric_fact.csv"
    aggr_metric_filename = f"{outdir}/{stem}.gsmm2_aligned_metric_aggr.csv"
    no_aligned_fact_filename = f"{outdir}/{stem}.non_aligned_fact.csv"
    no_aligned_aggr_filename = f"{outdir}/{stem}.non_aligned_aggr.csv"

    """
    fact_metric_filename = generate_metric_file(
        bam_file,
        ref_fa,
        out_filename=fact_metric_filename,
        force=force,
        threads=threads,
    )
    """

    processes = []
    if ref_fa != "" and enable_align:
        aligned_fact_thread = threading.Thread(target=generate_aligned_metric_fact_file, args=(
            bam_file, ref_fa, fact_metric_filename, short_aln,force, threads))
        aligned_fact_thread.start()
        processes.append(aligned_fact_thread)

    if enable_basic:
        non_aligned_fact_thread = threading.Thread(target=generate_non_aligned_metric_fact_file, args=(
            bam_file, no_aligned_fact_filename, outdir, force))
        non_aligned_fact_thread.start()
        processes.append(non_aligned_fact_thread)

    for p in processes:
        p.join()
    if ref_fa != "" and enable_align:
        aligned_metric_analysis(fact_metric_filename,
                                aggr_metric_filename, force=force)
    if enable_basic:
        non_aligned_metric_analysis(
            no_aligned_fact_filename, no_aligned_aggr_filename, force, out_dir=outdir, stem=stem)

    return (aggr_metric_filename, fact_metric_filename, no_aligned_aggr_filename, no_aligned_fact_filename)


def test_stat():
    fact_bam_basic = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/fact_aligned_bam_bam_basic.csv"
    aggr_metric_filename = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/aggr_metric.csv"

    with open(aggr_metric_filename, encoding="utf8", mode="w") as file_h:
        file_h.write(f"name\tvalue\n")
        stats(fact_bam_basic, file_h=file_h)


def expand_bam_files(bam_files):
    final_bam_files = []
    for bam_file in bam_files:
        if "*" in bam_file:
            final_bam_files.extend(glob(bam_file))
        else:
            final_bam_files.append(bam_file)
    return final_bam_files


def main_cli():
    """
    aligned bam analysis & origin bam analysis
    在 metric 中使用
    """

    parser = argparse.ArgumentParser(prog="parser")
    parser.add_argument("--bams", nargs="+", type=str,
                        required=True, help="wildcard '*' is supported")
    parser.add_argument("--refs", nargs="+", type=str,
                        help="if not provided. the alignment related metric will not output")
    parser.add_argument("--short-aln", type=int, default=0,
                        help="for query or target in [30, 200]", dest="short_aln")
    parser.add_argument(
        "-f",
        action="store_true",
        default=False,
        help="regenerate the metric file if exists",
    )
    args = parser.parse_args()

    bam_files = args.bams
    bam_files = expand_bam_files(bam_files)

    refs = args.refs

    if refs is None or len(refs) == 0:
        refs = [""] * len(bam_files)

    if len(refs) == 1:
        refs = refs * len(bam_files)

    assert len(bam_files) == len(refs)

    for bam, ref in zip(bam_files, refs):
        main(bam_file=bam, ref_fa=ref, short_aln=args.short_aln==1 ,force=args.f)


if __name__ == "__main__":
    main_cli()



import pysam
from tqdm import tqdm
import os
import argparse
import numpy as np
from multiprocessing import cpu_count
import matplotlib.pyplot as plt
import seaborn as sns
import pathlib
from typing import Dict, List, Tuple


def extract_multi_rq_length_counts(
    bam_file: str,
    channel_tag: str,
    rq_thresholds: List[float]
) -> Dict[float, Dict[int, int]]:
    """
    单次遍历 BAM，同时统计多个 rq 阈值的长度分布

    返回: {threshold: {length: count}}
    """
    assert channel_tag in ("ch", "zm")
    # 按阈值降序排序（优化：高阈值先检查，但无法短路因阈值非连续）
    thresholds_sorted = sorted(set(rq_thresholds), reverse=True)
    accumulators = {f"{t:.5f}": {} for t in thresholds_sorted}

    with pysam.AlignmentFile(
        filename=bam_file, mode="rb", check_sq=False, threads=cpu_count() // 2
    ) as reader:
        desc = f"Processing {os.path.basename(bam_file)} for {len(thresholds_sorted)} rq thresholds"
        for record in tqdm(reader.fetch(until_eof=True), desc=desc):
            # 计算 rq 值
            if record.has_tag("rq"):
                rq = float(record.get_tag("rq"))
            elif record.has_tag("cq"):
                rq = 1 - 10 ** (float(record.get_tag("cq")) / -10)
            else:
                rq = 1.0  # 无质量标签视为完美质量

            # 获取序列长度
            length = len(record.query_sequence)

            # 为所有满足条件的阈值累加计数
            for threshold in thresholds_sorted:
                if rq >= threshold:
                    threshold = f"{threshold:.5f}"
                    accumulators[threshold][length] = accumulators[threshold].get(
                        length, 0) + 1

    return accumulators


def weighted_median(values: np.ndarray, weights: np.ndarray) -> float:
    """安全计算加权中位数，避免内存爆炸"""
    sorted_idx = np.argsort(values)
    sorted_vals = values[sorted_idx]
    sorted_weights = weights[sorted_idx]
    cumsum = np.cumsum(sorted_weights)
    cutoff = cumsum[-1] / 2.0
    return float(sorted_vals[np.searchsorted(cumsum, cutoff)])


def plot_single_threshold(
    length_counts: Dict[int, int],
    bam_file: str,
    min_rq: float,
    output_file: str,
    bins: int = 1000,
    figsize: tuple = (10, 8),
    dpi: int = 300
):
    """单阈值分布：双子图（线性+对数坐标）"""
    if not length_counts:
        raise ValueError("未提取到任何有效序列，请检查 BAM 文件或过滤参数")

    # 转换为数组
    lengths = np.array(list(length_counts.keys()))
    counts = np.array(list(length_counts.values()))

    # 计算统计量
    total_reads = counts.sum()
    mean_len = np.average(lengths, weights=counts)
    median_len = weighted_median(lengths, counts)

    # 创建分箱
    min_len, max_len = lengths.min(), lengths.max()
    bins_linear = np.linspace(min_len, max_len, bins + 1)
    bins_log = np.logspace(np.log10(max(1, min_len)),
                           np.log10(max_len), bins + 1)

    # 绘图
    sns.set_theme(style="whitegrid", font_scale=1.2)
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize, sharey=False)

    # 线性坐标
    sns.histplot(
        x=lengths, weights=counts, bins=bins_linear,
        ax=ax1, color="steelblue", edgecolor="white", linewidth=0.5
    )
    ax1.set_title(f"Sequence Length Distribution (Linear Scale)\n"
                  f"BAM: {os.path.basename(bam_file)} | rq ≥ {min_rq}\n"
                  f"Total reads: {total_reads:,} | Mean: {mean_len:.1f} bp | Median: {median_len:.1f} bp",
                  fontsize=13, fontweight="bold")
    ax1.set_xlabel("Sequence Length (bp)", fontsize=12)
    ax1.set_ylabel("Read Count", fontsize=12)
    ax1.ticklabel_format(style='scientific', axis='y', scilimits=(0, 0))

    # 对数坐标
    sns.histplot(
        x=lengths, weights=counts, bins=bins_log,
        ax=ax2, color="coral", edgecolor="white", linewidth=0.5
    )
    ax2.set_xscale('log')
    ax2.set_title("Sequence Length Distribution (Log Scale)",
                  fontsize=13, fontweight="bold")
    ax2.set_xlabel("Sequence Length (bp, log scale)", fontsize=12)
    ax2.set_ylabel("Read Count", fontsize=12)
    ax2.ticklabel_format(style='scientific', axis='y', scilimits=(0, 0))

    plt.tight_layout()
    _save_figure(fig, output_file, dpi)

    print(f"✅ 单阈值分布图已保存: {os.path.abspath(output_file)}")
    print(
        f"📊 统计: rq≥{min_rq} | Reads={total_reads:,} | Mean={mean_len:.1f} bp | Median={median_len:.1f} bp")


def plot_multi_threshold(
    multi_counts: Dict[float, Dict[int, int]],
    rq_thresholds: List[float],
    output_file: str,
    bins: int = 100,
    figsize: tuple = (12, 8),
    dpi: int = 300
):
    """多阈值分布：单图叠加（对数坐标）"""
    # 按阈值降序排列（高质量在前）
    thresholds = sorted(rq_thresholds, reverse=False)

    # 全局长度范围（用于统一分箱）

    # 创建图形
    sns.set_theme(style="whitegrid", font_scale=1.2)
    fig, axs = plt.subplots(figsize=figsize, nrows=len(thresholds), ncols=1)

    # 配色方案（从高质量到低质量）
    colors = sns.color_palette("husl", len(thresholds))

    # 绘制每个阈值的分布
    stats_summary = []
    for idx, threshold in enumerate(thresholds):

        ax = axs[idx]
        counts_dict = multi_counts.get(f"{threshold:.5f}", {})
        if not counts_dict:
            print(f"⚠️  阈值 rq≥{threshold} 无有效数据，跳过")
            continue

        lengths = np.array(list(counts_dict.keys()))
        counts = np.array(list(counts_dict.values()))

        # ✅ 每个 subplot 自己的 x 轴范围（基于 length）
        cur_min = max(1, lengths.min())
        cur_max = lengths.max()
        bins_linear = np.linspace(cur_min, cur_max, bins + 1)

        # 统计量
        total = counts.sum()
        mean = np.average(lengths, weights=counts)
        median = weighted_median(lengths, counts)
        stats_summary.append((threshold, total, mean, median))

        # 画图
        ax.hist(
            lengths,
            bins=bins_linear,
            weights=counts,
            histtype='step',
            linewidth=2.0,
            color=colors[idx],
            alpha=0.9
        )

        # ===== 字体 & 轴设置 =====
        ax.set_ylabel(
            "Read Count",
            fontsize=9,
            fontweight="bold"
        )

        # 子图标题（小而清晰）
        ax.text(
            0.02, 0.95,
            f"rq ≥ {threshold:.3f} | n = {total:,}",
            transform=ax.transAxes,
            fontsize=5,
            fontweight="bold",
            va="top",
            ha="left"
        )

        ax.set_xlim(cur_min, cur_max)

        ax.tick_params(
            axis="both",
            which="major",
            labelsize=8
        )

        ax.grid(
            True,
            which="both",
            linestyle="--",
            linewidth=0.6,
            alpha=0.6
        )

        # ✅ 只有最后一个 subplot 显示 x label
        if idx == len(thresholds) - 1:
            ax.set_xlabel(
                "Sequence Length (bp)",
                fontsize=10,
                fontweight="bold"
            )
        else:
            ax.set_xlabel("")

    # plt.tight_layout()
    _save_figure(fig, output_file, dpi)

    # 打印统计摘要
    print(f"\n✅ 多阈值分布图已保存: {os.path.abspath(output_file)}")
    print("\n📊 各阈值统计摘要:")
    print(f"{'Threshold':<12} {'Total Reads':<15} {'Mean Length':<15} {'Median Length':<15}")
    print("-" * 60)
    for threshold, total, mean, median in stats_summary:
        print(
            f"rq ≥ {threshold:<5.2f} {total:>14,} {mean:>14.0f} bp {median:>14.0f} bp")


def _save_figure(fig: plt.Figure, output_file: str, dpi: int):
    """保存图形的辅助函数"""
    output_dir = os.path.dirname(os.path.abspath(output_file))
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)
    fig.savefig(output_file, dpi=dpi, bbox_inches="tight")
    plt.close(fig)


def generate_output_path(bam_path: str, rq_values: List[float], is_multi: bool) -> str:
    """生成输出文件路径"""
    p = pathlib.Path(bam_path)
    if is_multi:
        # 多阈值：用下划线连接阈值（保留2位小数）
        rq_str = "_".join(
            [f"{v:.2f}" for v in sorted(rq_values, reverse=True)])
        return str(p.parent / f"{p.stem}.rq_{rq_str}.length_dist.png")
    else:
        # 单阈值：保留原命名风格
        rq_str = f"{rq_values[0]:.2f}" if rq_values[0] > 0 else "all"
        return str(p.parent / f"{p.stem}.rq_{rq_str}.length_dist.png")


def main(args):
    # 确保阈值列表非空
    if not args.min_rq:
        args.min_rq = [0.0]

    for bam_path in args.bams:
        # 单次提取所有阈值的分布
        multi_counts = extract_multi_rq_length_counts(
            bam_path,
            args.channel_tag,
            args.min_rq
        )

        # 根据阈值数量选择绘图模式
        is_multi = len(args.min_rq) > 1
        output_path = generate_output_path(bam_path, args.min_rq, is_multi)

        if is_multi:
            plot_multi_threshold(
                multi_counts,
                args.min_rq,
                output_path,
                bins=args.bins,
                figsize=(12, 8),
                dpi=args.dpi
            )
        else:
            # 单阈值：取第一个阈值的数据
            single_counts = multi_counts.get(args.min_rq[0], {})
            plot_single_threshold(
                single_counts,
                bam_path,
                args.min_rq[0],
                output_path,
                bins=args.bins,
                figsize=(10, 8),
                dpi=args.dpi
            )


def main_cli():
    parser = argparse.ArgumentParser(
        prog="bam-length-dist",
        description="BAM file sequence length distribution analysis with rq filtering"
    )
    parser.add_argument("bams", nargs="+", type=str,
                        help="One or more BAM file paths")
    parser.add_argument(
        "--min_rq",
        nargs="+",
        type=float,
        default=[0.0, 0.95, 0.99, 0.999],
        help="One or more rq thresholds (e.g., --min_rq 0.7 0.8 0.9). Default: 0.0 (no filtering)"
    )
    parser.add_argument(
        "--channel_tag",
        type=str,
        default="ch",
        choices=["ch", "zm"],
        help="Channel tag to validate (ch or zm). Default: ch"
    )
    parser.add_argument(
        "--bins",
        type=int,
        default=1000,
        help="Number of histogram bins. Default: 100 (multi-threshold) / 1000 (single-threshold)"
    )
    parser.add_argument(
        "--dpi",
        type=int,
        default=300,
        help="Output image DPI. Default: 300"
    )
    args = parser.parse_args()

    # 单阈值时使用更细的分箱
    if len(args.min_rq) == 1 and args.bins == 100:
        args.bins = 1000

    main(args)


if __name__ == "__main__":
    main_cli()



import subprocess
import pathlib
import pysam
from tqdm import tqdm
import argparse
from glob import glob
import multiprocessing as mp


class RecordInfo:
    def __init__(self, name, length, rq):
        self.name = name
        self.length = length
        self.rq = rq


def read_bam_file(bam_file: str, rq_thr: float):
    infos = {}
    # 使用 pysam 读取 BAM 及其 rq 标签
    with pysam.AlignmentFile(bam_file, mode="rb", check_sq=False, threads=mp.cpu_count()) as in_bam:
        for record in tqdm(in_bam.fetch(until_eof=True), desc=f"reading BAM: {bam_file}"):
            rq = 1.0
            if record.has_tag("rq"):
                rq = record.get_tag("rq")

            if rq < rq_thr:
                continue

            infos[record.query_name] = RecordInfo(
                record.query_name, length=len(record.query_sequence), rq=rq)
    return infos


def process_fastx_to_fasta(input_file: str, output_fasta: str):
    """
    使用 pysam.FastxFile 读取 FASTQ 或 FASTA，
    并提取长度信息，同时写出一个标准的 FASTA 文件。
    """
    infos = {}
    with pysam.FastxFile(input_file) as fx:
        out_fa = None
        if input_file.endswith(".fq") or input_file.endswith(".fastq"):
            out_fa = open(output_fasta, "w")

        for entry in tqdm(fx, desc=f"processing {input_file}"):
            # 记录元数据
            infos[entry.name] = RecordInfo(
                entry.name, length=len(entry.sequence), rq=1.0)
            # 写入 FASTA 格式
            if out_fa is not None:
                out_fa.write(f">{entry.name}\n{entry.sequence}\n")
        if out_fa is not None:
            out_fa.flush()
            out_fa.close()

    return infos


def gff_reader(fname):
    infos = {}
    if not pathlib.Path(fname).exists():
        return infos
    with open(fname, mode="r", encoding="utf8") as in_file:
        for line in tqdm(in_file, desc=f"reading GFF: {fname}"):
            line = line.strip()
            if line.startswith("#") or not line:
                continue
            items = line.split("\t")
            if len(items) < 5:
                continue
            key = items[0]
            start = int(items[3])
            end = int(items[4])
            length = end - start + 1
            infos.setdefault(key, 0)
            infos[key] += length
    return infos


def main(args):
    input_files = []
    for f_pat in args.files:
        input_files.extend(glob(f_pat))

    if not input_files:
        print("No files found. Please check your input paths.")
        return

    print(f"Processing files: {input_files}")
    report_inner = ""

    for f_path in input_files:
        p = pathlib.Path(f_path)
        suffix = p.suffix.lower()
        fasta_file = p.parent.joinpath(f"{p.stem}.fasta")
        gff_file = p.parent.joinpath(f"{p.stem}.gff")

        # 1. 根据不同类型读取信息并统一生成/准备 FASTA
        if suffix == ".bam":
            record_infos = read_bam_file(bam_file=f_path, rq_thr=args.rq_thr)
            cmd_conv = f"samtools fasta {f_path} > {fasta_file}"
            print(f"running {cmd_conv}")
            subprocess.check_call(cmd_conv, shell=True)
        elif suffix in [".fastq", ".fq", ".fasta", ".fa"]:
            # 使用 pysam.FastxFile 处理，不再调用 shell 命令转换
            record_infos = process_fastx_to_fasta(f_path, str(fasta_file))
        else:
            print(f"Skipping unsupported file type: {suffix}")
            continue

        # 2. 索引 FASTA (tr-finder 通常需要 faidx)
        cmd_idx = f"samtools faidx {fasta_file}"
        print(f"running {cmd_idx}")
        subprocess.check_call(cmd_idx, shell=True)

        # 3. 运行 tr-finder
        cmd_tr = f"tr-finder called {fasta_file} --unitAndRepeats 1-4,2-3,3-3 -o {gff_file}"
        print(f"running {cmd_tr}")
        subprocess.check_call(cmd_tr, shell=True)

        # 4. 统计结果
        gff_infos = gff_reader(gff_file)

        tot_len = 0
        tr_len = 0
        for record_key, record_info in record_infos.items():
            tot_len += record_info.length
            tr_len += gff_infos.get(record_key, 0)

        ratio = (tr_len / tot_len * 100) if tot_len > 0 else 0
        res_line = f"File: {p.name} | Ratio: {ratio:.2f} %"
        report_inner += f"- {res_line}\n"

    report_str = f"""
================= HomoAndStr Report =================
{report_inner}
=====================================================
"""
    print(report_str)
    return report_str


def main_cli():
    parser = argparse.ArgumentParser(
        description="HomoAndStr Ratio Tool (Supports BAM/FASTQ/FASTA)")
    parser.add_argument("files", nargs="+",
                        help="Input files (.bam, .fq, .fastq, .fa, .fasta)")
    parser.add_argument('--rq-thr', type=float, default=0.95,
                        help="Read Quality threshold (BAM only)")
    args = parser.parse_args()
    main(args=args)


if __name__ == "__main__":
    main_cli()



import subprocess
import pathlib
import os
import logging
import polars as pl
import shutil
import argparse
from multiprocessing import cpu_count
import os
import semver


logging.basicConfig(
    level=logging.INFO,
    datefmt="%Y/%m/%d %H:%M:%S",
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def mm2_version_check():
    oup = subprocess.getoutput("gsmm2-time-err -V")
    oup = oup.strip()
    version_str = oup.rsplit(" ", maxsplit=1)[1]

    logging.info(f"gsmm2-time-err Version: {version_str}")
    mm2_version = semver.Version.parse(version_str)
    expected_version = "0.25.1"
    assert mm2_version >= semver.Version.parse(
        expected_version
    ), f"current mm2 version:{mm2_version} < {expected_version}, try 'cargo uninstall mm2; cargo install mm2@={expected_version}' "


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "300"
    os.environ["POLARS_FMT_STR_LEN"] = "100"


def extract_filename(filepath: str) -> str:
    p = pathlib.Path(filepath)
    return p.stem


def generate_metric_file(
    bam_file: str,
    ref_fasta: str,
    out_filename: str,
    force: bool = False,
    threads=None,
    no_supp=False,
    no_mar=False,
    short_aln=False,
) -> str:

    if no_supp and no_mar:
        raise ValueError("no_supp, no_mar can't be all true")

    if force and os.path.exists(out_filename):
        os.remove(out_filename)

    if os.path.exists(out_filename):
        logging.warning("use the existing metric file : %s", out_filename)
        return out_filename

    threads = cpu_count() if threads is None else threads
    cmd = f"""gsmm2-time-err --threads {threads} \
            -q {bam_file} \
            -t {ref_fasta} \
            --out {out_filename} \
            --kmer 11 \
            --wins 1 """
    if short_aln:
        cmd += " --short-aln"

    logging.info("cmd: %s", cmd)
    subprocess.check_call(cmd, shell=True)

    return out_filename


def stats(metric_filename, filename):
    # schema_overrides={"longIndel": pl.String}
    df = pl.read_csv(metric_filename, separator="\t")
    df = df.filter(pl.col("rname") != "")

    over_all_metrics = (
        df.select(
            [
                pl.col("eq").sum(),
                pl.col("diff").sum(),
                pl.col("ins").sum(),
                pl.col("del").sum(),
            ]
        )
        .with_columns(
            [
                (pl.col("eq") + pl.col("diff") + pl.col("ins") + pl.col("del")).alias(
                    "aligned_span"
                )
            ]
        )
        .select(
            [
                (pl.col("eq") / pl.col("aligned_span")).alias("eq_rate"),
                (pl.col("diff") / pl.col("aligned_span")).alias("diff_rate"),
                (pl.col("ins") / pl.col("aligned_span")).alias("ins_rate"),
                (pl.col("del") / pl.col("aligned_span")).alias("del_rate"),
            ]
        )
    )
    print(over_all_metrics)

    aggr_metrics = (
        df.group_by("minutes")
        .agg(
            [
                pl.col("eq").sum(),
                pl.col("diff").sum(),
                pl.col("ins").sum(),
                pl.col("del").sum(),
            ]
        )
        .with_columns(
            [
                (pl.col("eq") + pl.col("diff") + pl.col("ins") + pl.col("del")).alias(
                    "aligned_span"
                )
            ]
        )
        .with_columns(
            [
                (pl.col("eq") / pl.col("aligned_span")).alias("eq_rate"),
                (pl.col("diff") / pl.col("aligned_span")).alias("diff_rate"),
                (pl.col("ins") / pl.col("aligned_span")).alias("ins_rate"),
                (pl.col("del") / pl.col("aligned_span")).alias("del_rate"),
            ]
        )
        .drop(["eq", "diff", "ins", "del"])
        .sort(by=["minutes"], descending=[False])
    )

    print(aggr_metrics)

    aggr_metrics.write_csv(filename, include_header=True, separator="\t")


def stats_v2(metric_filename, filename):
    # schema_overrides={"longIndel": pl.String}
    df = pl.read_csv(metric_filename, separator="\t")
    df = df.filter(pl.col("rname") != "")
    distinct_life_times = (
        df.group_by("minutes")
        .agg([pl.len()])
        .select([pl.col("minutes")])
        .to_pandas()["minutes"]
        .to_list()
    )
    qname_life_time = df.group_by("qname").agg([pl.col("minutes").max()])

    aggr_metrics = []

    over_all_metrics = (
        df.select(
            [
                pl.col("eq").sum(),
                pl.col("diff").sum(),
                pl.col("ins").sum(),
                pl.col("del").sum(),
            ]
        )
        .with_columns(
            [
                (pl.col("eq") + pl.col("diff") + pl.col("ins") + pl.col("del")).alias(
                    "aligned_span"
                )
            ]
        )
        .select(
            [
                (pl.col("eq") / pl.col("aligned_span")).alias("eq_rate"),
                (pl.col("diff") / pl.col("aligned_span")).alias("diff_rate"),
                (pl.col("ins") / pl.col("aligned_span")).alias("ins_rate"),
                (pl.col("del") / pl.col("aligned_span")).alias("del_rate"),
            ]
        )
    )
    print(over_all_metrics)

    for cur_life_time in sorted(distinct_life_times):
        interested_qname = qname_life_time.filter(
            pl.col("minutes").eq(pl.lit(cur_life_time))
        ).select([pl.col("qname")])

        aggr_metric = (
            df.join(interested_qname, on="qname", how="inner")
            .group_by("minutes")
            .agg(
                [
                    pl.col("eq").sum(),
                    pl.col("diff").sum(),
                    pl.col("ins").sum(),
                    pl.col("del").sum(),
                ]
            )
            .with_columns(
                [
                    (
                        pl.col("eq") + pl.col("diff") + pl.col("ins") + pl.col("del")
                    ).alias("aligned_span")
                ]
            )
            .with_columns(
                [
                    (pl.col("eq") / pl.col("aligned_span")).alias("eq_rate"),
                    (pl.col("diff") / pl.col("aligned_span")).alias("diff_rate"),
                    (pl.col("ins") / pl.col("aligned_span")).alias("ins_rate"),
                    (pl.col("del") / pl.col("aligned_span")).alias("del_rate"),
                ]
            )
            .drop(["eq", "diff", "ins", "del"])
            .sort(by=["minutes"], descending=[False])
            .with_columns(pl.lit(cur_life_time).alias("life_time"))
            .select(
                [
                    pl.col("life_time"),
                    pl.col("minutes"),
                    pl.col("aligned_span"),
                    pl.col("eq_rate"),
                    pl.col("diff_rate"),
                    pl.col("ins_rate"),
                    pl.col("del_rate"),
                ]
            )
        )
        aggr_metrics.append(aggr_metric)

    aggr_metrics = pl.concat(aggr_metrics)

    print(aggr_metrics)

    aggr_metrics.write_csv(filename, include_header=True, separator="\t")


def main(
    bam_file: str,
    ref_fa: str,
    threads=None,
    force=False,
    short_aln=False,
    outdir=None,
    copy_bam_file=False,
) -> str:
    """
        step1: generate detailed metric info
        step2: compute the aggr metric. the result aggr_metric.csv is a '\t' seperated csv file. the header is name\tvalue
            here is a demo.
            ---------aggr_metric.csv
            name    value
            queryCoverage   0.937
            ----------

    requirements:
        mm2: cargo install mm2 (version >= 0.19.0)

    Args:
        bam_file (str): bam file. only support adapter.bam
        ref_fa (str): ref genome fa file nam
        threads (int|None): threads for generating detailed metric file
        force (boolean): if force==False, use the existing metric file if exists
        outdir: if None, ${bam_filedir}/${bam_file_stem}-metric as outdir
        copy_bam_file: copy bam file to outdir. Set this parameter to true when the file is on the NAS.

    Return:
        (aggr_metric_filename, fact_metric_filename) (str, str)
    """

    mm2_version_check()

    if copy_bam_file:
        assert outdir is not None, "must provide outdir when copy_bam_file=True"
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        new_bam_file = os.path.join(outdir, os.path.basename(bam_file))
        if os.path.exists(new_bam_file):
            raise ValueError(f"{new_bam_file} already exists")
        shutil.copy(bam_file, new_bam_file)
        bam_file = new_bam_file

    bam_filedir = os.path.dirname(bam_file)
    stem = extract_filename(bam_file)
    if outdir is None:
        outdir = os.path.join(bam_filedir, f"{stem}-metric")

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    fact_metric_filename = f"{outdir}/{stem}.gsmm2-time-err.csv"
    fact_metric_filename = generate_metric_file(
        bam_file,
        ref_fa,
        out_filename=fact_metric_filename,
        force=force,
        threads=threads,
        short_aln=short_aln,
    )
    aggr_metric_filename = f"{outdir}/{stem}.gsmm2-time-err-aggr.csv"
    if force and os.path.exists(aggr_metric_filename):
        os.remove(aggr_metric_filename)

    # if not os.path.exists(aggr_metric_filename):
    stats_v2(fact_metric_filename, filename=aggr_metric_filename)
    # else:
    #     logging.warning(
    #         "aggr_metric_file exists, use existing one. %s", aggr_metric_filename
    #     )
    return (aggr_metric_filename, fact_metric_filename)


def test_stat():
    fact_bam_basic = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/fact_aligned_bam_bam_basic.csv"
    aggr_metric_filename = "/data/adapter-query-coverage-valid-data/20250107_240901Y0007_Run0001_adapter-metric/metric/aggr_metric.csv"

    with open(aggr_metric_filename, encoding="utf8", mode="w") as file_h:
        stats(fact_bam_basic, file_h=file_h)


def main_cli():
    polars_env_init()

    parser = argparse.ArgumentParser(prog="parser")
    parser.add_argument("--bams", nargs="+", type=str, required=True)
    parser.add_argument("--refs", nargs="+", type=str, required=True)
    parser.add_argument(
        "--short-aln",
        type=int,
        default=0,
        help="for query or target in [30, 200]",
        dest="short_aln",
    )
    parser.add_argument(
        "-f",
        action="store_true",
        default=False,
        help="regenerate the metric file if exists",
    )
    args = parser.parse_args()

    bam_files = args.bams
    refs = args.refs
    if len(refs) == 1:
        refs = refs * len(bam_files)

    assert len(bam_files) == len(refs)

    for bam, ref in zip(bam_files, refs):
        main(bam_file=bam, ref_fa=ref, force=args.f, short_aln=args.short_aln == 1)


if __name__ == "__main__":
    main_cli()



import re


def revcomp(seq):
    table = str.maketrans("ACGTacgt", "TGCAtgca")
    return seq.translate(table)[::-1]


def parse_cigar(cigar):
    """把 CIGAR 拆成 [(len, op), ...]"""
    num = ""
    for c in cigar:
        if c.isdigit():
            num += c
        else:
            yield int(num), c
            num = ""


def blast_like_alignment(query, ref, aln, line_width=110):
    q_pos = aln.q_st
    if aln.strand == -1:
        q_pos = len(query) - aln.q_en
        query = revcomp(query)
    r_pos = aln.r_st

    q_line = []
    r_line = []
    mid_line = []

    for length, op in parse_cigar(aln.cigar_str):
        if op == "M" or op == "=" or op == "X":
            for i in range(length):
                qb = query[q_pos]
                rb = ref[r_pos]
                q_line.append(qb)
                r_line.append(rb)
                mid_line.append("|" if qb == rb else " ")
                q_pos += 1
                r_pos += 1

        elif op == "I":  # 插入到 reference
            for i in range(length):
                qb = query[q_pos]
                q_line.append(qb)
                r_line.append("-")
                mid_line.append(" ")
                q_pos += 1

        elif op == "D":  # reference 多出来
            for i in range(length):
                rb = ref[r_pos]
                q_line.append("-")
                r_line.append(rb)
                mid_line.append(" ")
                r_pos += 1

        elif op == "S":  # soft clip，通常忽略
            q_pos += length

    # 分段打印（像 blastn）
    q_str = "".join(q_line)
    r_str = "".join(r_line)
    m_str = "".join(mid_line)

    for i in range(0, len(q_str), line_width):
        q_seg = q_str[i:i+line_width]
        m_seg = m_str[i:i+line_width]
        r_seg = r_str[i:i+line_width]

        q_start = aln.q_st + i + 1
        q_end = q_start + len(q_seg) - 1
        r_start = aln.r_st + i + 1
        r_end = r_start + len(r_seg.replace("-", "")) - 1

        print(f"Sbjct  {r_start:<6} {r_seg} {r_end}")
        print(f"              {m_seg}")
        print(f"Query  {q_start:<6} {q_seg} {q_end}")

        print()


def calculate_identity_from_cigar(cigar_string):
    pattern = r'(\d+)([=IDX])'
    matches = re.findall(pattern, cigar_string)

    match_count = 0
    mismatch_count = 0
    total_aligned = 0

    for length_str, operation in matches:
        length = int(length_str)
        if operation == '=':
            match_count += length
            total_aligned += length
        elif operation == 'X':
            mismatch_count += length
            total_aligned += length
        elif operation in ('I', 'D'):
            total_aligned += length

    if total_aligned == 0:
        return 0.0
    return match_count / total_aligned



from pathlib import Path


def extract_chr9(input_fasta: str, output_fasta: str):
    with open(input_fasta, "r") as infile, open(output_fasta, "w") as outfile:
        write = False
        for line in infile:
            if line.startswith(">"):
                if "chr9" in line:
                    write = True
                    outfile.write(line)  # 写入 fasta header
                else:
                    write = False
            elif write:
                outfile.write(line)


# 示例调用
extract_chr9(
    input_fasta="/data/ccs_data/HG002/GCA_000001405.15_GRCh38_no_alt_analysis_set.chr1-chr22.fasta",
    output_fasta="/data/REF_GENOMES/chr9.fasta",
)



import os
import argparse
import time
import mappy
import re
import pysam
from multiprocessing import Process, Queue, Value, Lock, cpu_count
import signal
import sys

# =========================
# 全局计数器（多进程安全）
# =========================
TOTAL_READS = Value('i', 0)
SPLIT_READS = Value('i', 0)
COUNTER_LOCK = Lock()

# Sentinel for queue termination
_SENTINEL = None


def calculate_identity_from_cigar(cigar_string):
    pattern = r'(\d+)([=IDX])'
    matches = re.findall(pattern, cigar_string)

    match_count = 0
    mismatch_count = 0
    total_aligned = 0

    for length_str, operation in matches:
        length = int(length_str)
        if operation == '=':
            match_count += length
            total_aligned += length
        elif operation == 'X':
            mismatch_count += length
            total_aligned += length
        elif operation in ('I', 'D'):
            total_aligned += length

    if total_aligned == 0:
        return 0.0
    return match_count / total_aligned


def try_split_seq_2(seq: str):
    seq_len = len(seq)
    if seq_len < 200 or seq_len > 20000:
        return False

    mid = seq_len // 2
    seq1 = seq[:mid]
    seq2 = seq[mid:]

    aligner = mappy.Aligner(seq=seq1, extra_flags=67108864,
                            k=9, w=7, best_n=10, n_threads=1)
    for hit in aligner.map(seq2):
        identity = calculate_identity_from_cigar(hit.cigar_str)
        coverage1 = (hit.r_en - hit.r_st) / len(seq1)
        coverage2 = (hit.q_en - hit.q_st) / len(seq2)
        if identity > 0.80 and coverage1 > 0.85 and coverage2 > 0.85:
            return True
    return False


# =========================
# 消费者 Worker
# =========================
def consumer(queue):
    local_total = 0
    local_split = 0

    while True:
        seq = queue.get()
        if seq is _SENTINEL:
            # Put sentinel back for other consumers
            queue.put(_SENTINEL)
            break

        if seq and len(seq) > 0:
            local_total += 1
            if try_split_seq_2(seq):
                local_split += 1

    # Update global counters
    with COUNTER_LOCK:
        TOTAL_READS.value += local_total
        SPLIT_READS.value += local_split


# =========================
# 生产者：从 BAM 读取并放入队列
# =========================
def producer(bam_path, queue):
    count = 0
    with pysam.AlignmentFile(bam_path, "rb", threads=10, check_sq=False) as bam:
        for read in bam.fetch(until_eof=True):
            seq = read.query_sequence
            if seq:
                queue.put(seq)
                count += 1
                if count % 10000 == 0:
                    print(f"\rProduced {count} reads...",
                          end='', flush=True)
    print(f"\nProducer finished. Total reads enqueued: {count}")


# =========================
# 主函数
# =========================
def main(bam_path, threads):
    start = time.time()

    # 创建队列（适当大小避免爆内存）
    queue = Queue(maxsize=threads * 10)

    # 启动消费者进程
    consumers = []
    for _ in range(threads):
        p = Process(target=consumer, args=(queue,))
        p.start()
        consumers.append(p)

    # 启动生产者（在主进程中运行，简化控制）
    print("Starting producer...")
    producer(bam_path, queue)

    # 发送终止信号
    queue.put(_SENTINEL)

    # 等待所有消费者结束
    for p in consumers:
        p.join()

    # 输出结果
    total = TOTAL_READS.value
    split = SPLIT_READS.value
    ratio = split / total if total > 0 else 0.0
    report_str = f"""
================= Macebell Report =================
- Reads Processed : {total}
- Macebell Reads  : {split}
- Macebell Ratio  : {ratio*100: .2f}%
- Elapsed Time    : {time.time() - start:.2f}s
=======================================================
"""
    print(report_str)
    return report_str


def main_cli():
    parser = argparse.ArgumentParser(
        description="Estimate chimeric read split ratio from BAM using producer-consumer model."
    )
    parser.add_argument(
        "-i", "--input-bam",
        required=True,
        help="Input BAM file"
    )
    parser.add_argument(
        "-p", "--threads",
        type=int,
        default=cpu_count(),
        help="Number of consumer processes (default: all CPUs)"
    )
    args = parser.parse_args()

    if not os.path.exists(args.input_bam):
        raise FileNotFoundError(f"BAM file not found: {args.input_bam}")

    # Handle Ctrl+C gracefully
    def signal_handler(sig, frame):
        print("\nInterrupted by user. Exiting...")
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    main(args.input_bam, args.threads)


if __name__ == "__main__":
    main_cli()



import pysam
import argparse
from tqdm import tqdm


def bam_to_fastx(bam_path, fastx_path: str, rq_threshold):
    tot = 0
    dumped = 0
    output_fq = fastx_path.endswith("fastq")

    with pysam.AlignmentFile(
        bam_path, "rb", threads=40, check_sq=False
    ) as bam_file, open(fastx_path, "w") as fastx_out:
        for read in tqdm(
            bam_file.fetch(until_eof=True), desc=f"dumping {bam_path} to {fastx_path}"
        ):
            tot += 1

            # 尝试获取 rq 字段
            try:
                rq = read.get_tag("rq")
                if rq < rq_threshold:
                    continue
            except KeyError:
                pass

            # 构造 FASTQ 格式
            name = read.query_name
            seq = read.query_sequence
            qual = read.qual  # 转换为 ASCII 的质量字符串

            dumped += 1

            if seq is None or qual is None:
                continue  # 有可能 read 被软裁剪或缺失，跳过
            if output_fq:
                fastx_out.write(f"@{name}\n{seq}\n+\n{qual}\n")
            else:
                fastx_out.write(f">{name}\n{seq}\n")

    print(f"Tot:{tot}, dumped:{dumped}, ratio:{dumped / tot: .4f}")
    print(f"转换完成，输出文件: {fastx_path}")


def main_cli():
    parser = argparse.ArgumentParser(description="Convert BAM to FASTQ with rq filter.")
    parser.add_argument("bam", help="Input BAM file path")
    parser.add_argument("fastx", help="Output FASTX file path. fasta/fastq")
    parser.add_argument(
        "--rq", type=float, default=0.0, help="Minimum rq threshold (default: 0.0)"
    )

    args = parser.parse_args()
    assert args.fastx.endswith("fastq") or args.fastx.endswith("fasta"), "only fastq/fasta are supported"
    bam_to_fastx(args.bam, args.fastx, args.rq)


if __name__ == "__main__":

    main_cli()



import polars as pl
import math
import os
import argparse
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np


def set_polars_env():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "100"
    os.environ["POLARS_FMT_STR_LEN"] = "50"


def phreq2q(v):
    return 1 - math.pow(10, v / -10)


def q2phreqExpr(name, new_name=None):
    if new_name is None:
        new_name = name
    return (
        -10
        * (
            1
            - pl.when(pl.col(name) > (1 - 1e-10))
            .then(1 - 1e-10)
            .otherwise(pl.col(name))
        ).log10()
    ).alias(new_name)


def AccuracyExpr(phreq_threshold):
    threshold = phreq2q(phreq_threshold)
    # threshold = phredq_threshold
    return (
        pl.col("rq").ge(threshold).and_(pl.col("iy").ge(threshold)).sum()
        / pl.col("rq").ge(threshold).sum()
    ).alias(f"≥Q{phreq_threshold}Accuracy")


def RecallExpr(phreq_threshold):
    threshold = phreq2q(phreq_threshold)
    # threshold = phredq_threshold
    return (
        pl.col("rq").ge(threshold).and_(pl.col("iy").ge(threshold)).sum()
        / pl.col("iy").ge(threshold).sum()
    ).alias(f"≥Q{phreq_threshold}Recall")


def plot_rq_iy_scatter(df: pl.DataFrame, o_prefix=None):
    figure = plt.figure(figsize=(10, 10))
    axs = figure.add_subplot(1, 1, 1)
    plt.sca(axs)
    plt.grid(True, linestyle=":", linewidth=0.5, color="gray")

    df = df.to_pandas()
    sns.scatterplot(df, x="phreq_rq", y="phreq_iy", ax=axs)
    sns.kdeplot(
        df,
        x="phreq_rq",
        y="phreq_iy",
        ax=axs,
        cmap="RdYlGn",
        fill=True,
        alpha=0.5,
        label="Density",
        cbar=True,
    )

    axs.set_xticks(list(range(0, 60, 2)))
    axs.set_yticks(list(range(0, 60, 2)))
    axs.set_xlabel("PredictedChannelQ", fontdict={"size": 16})
    axs.set_ylabel("EmpericalChannelQ", fontdict={"size": 16})
    perfect_line = pl.DataFrame(
        {
            "x": list(range(0, 60)),
            "y": list(range(0, 60)),
        }
    )

    sns.lineplot(
        perfect_line.to_pandas(), x="x", y="y", ax=axs, color="blue", linestyle="--"
    )
    fname = "rq_iy_scatter.png"
    if o_prefix is not None:
        fname = f"{o_prefix}-{fname}"
    figure.savefig(fname=fname)
    

def plot_predQ20_but_lower_Q20(df: pl.DataFrame, o_prefix=None):
    figure = plt.figure(figsize=(10, 10))
    axs = figure.add_subplot(1, 1, 1)
    plt.sca(axs)
    plt.grid(True, linestyle=":", linewidth=0.5, color="gray")

    df = df.filter(pl.col("phreq_rq") >= pl.lit(20))\
        .filter(pl.col("phreq_iy") < pl.lit(20))\
        .select([pl.col("phreq_iy")])\
        .to_pandas()
    
    axs.set_xlim((10, 20))
    axs.set_xticks(np.linspace(10, 20, num=21))
    
    sns.histplot(df, x="phreq_iy", ax=axs, bins=100)

    axs.set_xlabel("TrueQ", fontdict={"size": 16})
    axs.set_ylabel("Count", fontdict={"size": 16})
    plt.title("True Q-score distribution for predicted Q >= 20")
    fname = "pred_lt_q20_but_lower_q20.png"
    if o_prefix is not None:
        fname = f"{o_prefix}-{fname}"
    figure.savefig(fname=fname)
    
    pass


def stat(fname: str, o_prefix):
    df = pl.read_csv(fname, separator="\t")
    df = df.with_columns(
        [
            pl.when(pl.col("rq") > 0.99999)
            .then(0.99999)
            .otherwise(pl.col("rq"))
            .alias("rq"),
            pl.when(pl.col("iy") > 0.99999)
            .then(0.99999)
            .otherwise(pl.col("iy"))
            .alias("iy"),
        ]
    ).with_columns([q2phreqExpr("rq", "phreq_rq"), q2phreqExpr("iy", "phreq_iy")])

    plot_rq_iy_scatter(df=df, o_prefix=o_prefix)
    plot_predQ20_but_lower_Q20(df=df, o_prefix=o_prefix)

    stat_res = df.select(
        (pl.col("rq") - pl.col("iy")).mean().alias("ME(pred-real)(rq-iy)"),
        (pl.col("rq") - pl.col("iy")).median().alias("MedE"),
        (pl.col("rq") - pl.col("iy")).abs().mean().alias("MAE"),
        ((pl.col("rq") - pl.col("iy")).abs() / pl.col("iy")).mean().alias("MAPE"),
        AccuracyExpr(13),
        RecallExpr(13),
        AccuracyExpr(20),
        RecallExpr(20),
        AccuracyExpr(25),
        RecallExpr(25),
        AccuracyExpr(30),
        RecallExpr(30),
    )
    print(stat_res)

    # .with_columns([pl.col("phreq_rq").mul(1/3).cast(pl.Int32).mul(3)])\
    stats2 = (
        df.with_columns([pl.col("phreq_rq").cast(pl.Int32)])
        .group_by("phreq_rq")
        .agg(
            [
                (pl.col("rq") - pl.col("iy")).mean().alias("ME(pred-real)(rq-iy)"),
                (pl.col("rq") - pl.col("iy")).median().alias("MedE"),
                (pl.col("rq") - pl.col("iy")).abs().mean().alias("MAE"),
                ((pl.col("rq") - pl.col("iy")).abs() / pl.col("iy"))
                .mean()
                .alias("MAPE"),
            ]
        )
        .sort(["phreq_rq"], descending=[False])
    )
    print(stats2)

    stats3 = (
        df.with_columns(
            [
                q2phreqExpr("rq", "phreq_rq").cast(pl.Int32),
            ]
        )
        .group_by("phreq_rq")
        .agg(
            [
                pl.col("iy").mean().alias("avg"),
                pl.col("iy").median().alias("median"),
                pl.col("iy").std().alias("std"),
                pl.col("phreq_iy").min().alias("min"),
                pl.quantile("phreq_iy", quantile=0.05).alias("percent_5"),
                pl.quantile("phreq_iy", quantile=0.20).alias("percent_20"),
                pl.quantile("phreq_iy", quantile=0.50).alias("percent_50"),
                pl.quantile("phreq_iy", quantile=0.80).alias("percent_80"),
                pl.quantile("phreq_iy", quantile=0.95).alias("percent_95"),
                pl.col("phreq_iy").max().alias("max"),
                pl.len().alias("cnt")
            ]
        )
        .with_columns(
            [
                q2phreqExpr("avg"),
                q2phreqExpr("median"),
            ]
        )
        .select(
            pl.col("phreq_rq"),
            pl.col("avg"),
            pl.col("median"),
            pl.concat_str(
                pl.col("min").round(2),
                pl.col("percent_5").round(2),
                pl.col("percent_20").round(2),
                pl.col("percent_50").round(2),
                pl.col("percent_80").round(2),
                pl.col("percent_95").round(2),
                pl.col("max").round(2),
                separator=", ",
            ).alias("Percent_0_5_20_50_80_95_100"),
            pl.col("cnt")
        )
        .sort(["phreq_rq"], descending=[False])
    )
    print(stats3)


def main(args):
    stat(args.fact_table, args.o_prefix)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("rq_iy_diff")
    parser.add_argument("fact_table", metavar="fact_aligned_bam_bam_basic")
    parser.add_argument("--o-prefix", metavar="o-prefix", default=None, dest="o_prefix")

    args_ = parser.parse_args()

    set_polars_env()
    main(args_)



from datetime import datetime
import re
import matplotlib.pyplot as plt
import glob
from typing import List, Tuple
import polars as pl
import os


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "30000"
    os.environ["POLARS_FMT_STR_LEN"] = "100"


def get_files(root_dir: str) -> List[str]:
    files = list(glob.glob(f"{root_dir}/*called-metric/*gsmm2-hp-aggr.csv"))
    return files


def read_datas(files: List[str]) -> List[Tuple[str, pl.DataFrame]]:
    result = []
    for fname in files:
        data = pl.read_csv(fname, separator="\t")
        result.append([fname, data])
    return result


def filter_data(data: pl.DataFrame) -> pl.DataFrame:
    return data.filter(pl.col("called") < (pl.col("true_cnt") + pl.lit(3))
                       ).filter(pl.col("called") > (pl.col("true_cnt") - pl.lit(3)))


def extract_date(filename: str) -> datetime.date:
    """从文件名提取日期(yyyymmdd格式)"""
    match = re.search(r"/(\d{8})_", filename)
    if match:
        return datetime.strptime(match.group(1), "%Y%m%d").date()
    raise ValueError(f"无法从文件名提取日期: {filename}")


def process_and_plot(data_list: List[Tuple[str, pl.DataFrame]], output_dir: str = "plots"):
    """主处理函数：合并数据并生成可视化图表"""
    # 数据预处理
    processed = []
    for filename, df in data_list:
        try:
            date = extract_date(filename)
            processed.append(df.with_columns(
                pl.lit(date).alias("date"),
                pl.col("called").cast(str)
            ))
        except ValueError as e:
            print(f"警告：{str(e)}")

    if not processed:
        print("无有效数据可处理")
        return

    combined = pl.concat(processed).sort("date")
    os.makedirs(output_dir, exist_ok=True)

    # 按true_cnt分组处理
    for cnt, cnt_df in combined.group_by("true_cnt"):
        fig, axes = plt.subplots(4, 2, figsize=(20, 15))
        fig.suptitle(f"True Count = {cnt}", fontsize=16)

        bases = ["A", "C", "G", "T"]
        tags = ["pure", "mixed"]

        # 绘制4x2子图矩阵
        for i, base in enumerate(bases):
            for j, tag in enumerate(tags):
                ax = axes[i][j]
                subset = cnt_df.filter(
                    (pl.col("true_base") == base) &
                    (pl.col("tag") == tag)
                )

                # 按called值分组绘制曲线
                for called, group in subset.group_by("called"):
                    sorted_df = group.sort("date")
                    ax.plot(
                        sorted_df["date"],
                        sorted_df["ratio_within_motif"],
                        marker="o",
                        label=f"called={called}"
                    )

                ax.set_title(f"Base {base} | Tag {tag}")
                ax.set_xlabel("Date")
                ax.set_ylabel("Ratio Within Motif")
                ax.legend()
                ax.grid(True)

        plt.tight_layout()
        output_path = os.path.join(output_dir, f"true_cnt_{cnt}.png")
        plt.savefig(output_path, dpi=300, bbox_inches="tight")
        plt.close()
        print(f"已生成图表: {output_path}")


def main():
    files = files = list(glob.glob("/data/metric-data/*smc_all_reads-metric/*gsmm2-hp-aggr.csv"))
    file_and_datas = read_datas(files)
    file_and_datas = [(v[0], filter_data(v[1])) for v in file_and_datas]
    file_and_datas = sorted(file_and_datas, key=lambda x: x[0])
    
    for file_and_data in file_and_datas:
        print(file_and_data[0])
    dedup_file_and_datas = []
    
    for file_and_data in file_and_datas:
        if len(dedup_file_and_datas) == 0:
            dedup_file_and_datas.append(file_and_data)
            continue
        date = extract_date(file_and_data[0])
        # assert isinstance(date, datetime.date)
        if dedup_file_and_datas[-1][0].rsplit("/", maxsplit=1)[-1].startswith(datetime.strftime(date, "%Y%m%d")):
            continue
        dedup_file_and_datas.append(file_and_data)
    print(len(file_and_datas), len(dedup_file_and_datas))
    process_and_plot(dedup_file_and_datas, output_dir="plotdir-sa-called")


if __name__ == "__main__":
    polars_env_init()
    main()



import os
import pysam
import logging
import argparse
from tqdm import tqdm

logging.basicConfig(
    level=logging.INFO,
    datefmt="%Y/%m/%d %H:%M:%S",
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def filter_called_bam(smc_bam, called_bam, output_bam):
    """
    对bam文件进行过滤
    :param smc_bam: smc之后的bam文件
    :param called_bam: basecaller之后的bam文件
    :param output_bam: 过滤之后的output bam文件
    :param temp_dir: 临时目录
    :return:
    """
    # 调用日志系统

    # 检查输入文件路径是否存在
    if not os.path.exists(smc_bam):
        logging.error(f"Input SMC BAM file {smc_bam} does not exist.")
        raise FileNotFoundError(
            f"Input SMC BAM file {smc_bam} does not exist.")  # 抛出异常

    if not os.path.exists(called_bam):
        logging.error(f"Input called BAM file {called_bam} does not exist.")
        raise FileNotFoundError(
            f"Input called BAM file {called_bam} does not exist.")  # 抛出异常

    smc_read_names = set()
    try:
        with pysam.AlignmentFile(smc_bam, "rb", check_sq=False, threads=os.cpu_count()) as a_bam:
            for read in tqdm(a_bam.fetch(until_eof=True), desc=f"reading {smc_bam}"):
                base_name = read.query_name.split()[0]
                # Adjust the parsing according to your read name format
                channel_number = base_name.split('/')[1]
                smc_read_names.add("read_" + channel_number)
        logging.info(
            f"Extracted {len(smc_read_names)} unique read names from {smc_bam}.")
    except Exception as e:
        logging.error(
            f"Failed to extract read names from {smc_bam}. Error: {e}")
        raise  # 重新抛出异常

    with pysam.AlignmentFile(filename=called_bam, mode="rb", check_sq=False, threads=os.cpu_count() // 2) as in_bam:
        with pysam.AlignmentFile(filename=output_bam, mode="wb", check_sq=False, threads=os.cpu_count() // 2, header=in_bam.header) as out_bam:
            cnt = 0
            for read in tqdm(in_bam.fetch(until_eof=True), desc=f"dumping {called_bam} to {output_bam}"):
                if read.query_name in smc_read_names:
                    cnt += 1 
                    out_bam.write(read)
            logging.info(f"dump {cnt} records")

def main(smc_bam: str, called_bam: str, output_bam: str):
    filter_called_bam(smc_bam=smc_bam, called_bam=called_bam,
                      output_bam=output_bam)



def main_cli():

    parser = argparse.ArgumentParser(prog="parser")
    parser.add_argument("--called-bam", type=str,
                        required=True, dest="called_bam")
    parser.add_argument("--smc-bam", type=str, required=True, dest="smc_bam")
    parser.add_argument("--out-bam", type=str, default=None, dest="out_bam")

    args = parser.parse_args()
    called_bam = args.called_bam
    smc_bam = args.smc_bam
    assert isinstance(called_bam, str)
    out_bam = args.out_bam
    if out_bam is None:
        out_bam = "{}.output.bam".format(called_bam.rsplit(".")[0])
    main(smc_bam, called_bam, out_bam)        

if __name__ == "__main__":
    main_cli()



import pysam
from tqdm import tqdm
import numpy as np

def q30_base_ratio_in_q30_channel(bam_file: str):
    q30_base_cnt = 0
    tot_base_cnt = 0
    with pysam.AlignmentFile(bam_file, mode="rb", check_sq=False) as bam_in:
        for record in tqdm(bam_in.fetch(until_eof=True), desc=f"reading {bam_file}"):
            rq = float(record.get_tag("rq"))
            if rq < 0.999:
                continue
            qual = record.query_qualities
            qual = np.array(qual)
            q30_base_cnt += (qual >= 30).sum()
            tot_base_cnt += len(qual)

    print(f"{bam_file} --> q30_base_ratio_in_q30_channel:{q30_base_cnt / tot_base_cnt * 100. :3f}%")
def main():
    fname = "/data1/ccs_data/20250804-ludaopei/20250721_240601Y0005_Run0001.polish2.smc_all_reads.bam"
    fname = "/data1/ccs_data/20250804-ludaopei/20250721_240601Y0005_Run0001.smc_all_reads.bam"
    fname = "/data1/ccs_data/20250804-ludaopei/20250721_240601Y0005_Run0001.smc_all_reads.bam"
    fnames = [
        
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002_adapter.bystrand.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.pplv4.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.pplv4.bystrand.qcali.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.polish3.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.pplv4.bystrand.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.pplv5.0.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002.pplv5.bystrand.smc_all_reads.bam",
        "/data1/ccs_data/20250804-ludaopei/20250804_240601Y0005_Run0002_adapter.xgb.smc_all_reads.bam",
        
    ]
    
    for fname in fnames:
        q30_base_ratio_in_q30_channel(fname)


if __name__ == "__main__":
    main()
    pass


import pysam
from typing import List, Set
import argparse
import pysam.samtools
from tqdm import tqdm


def extract_non_multiple_mapping_reads(bam_file: str) -> Set[str]:
    qname_cnts = {}
    with pysam.AlignmentFile(filename=bam_file, mode="rb", threads=40) as bam_h:
        for record in tqdm(
            bam_h.fetch(), desc=f"extract_non_multiple_mapping_reads from {bam_file}"
        ):
            qname = record.query_name
            if qname not in qname_cnts:
                qname_cnts[qname] = 0
            qname_cnts[qname] += 1
    return set([qname for qname, cnt in qname_cnts.items() if cnt == 1])


def dump_non_multiple_mapping_reads(bam_file: str, qname_set):
    o_bam_file = "{}.non_multiple_mapping_reads.bam".format(
        bam_file.rsplit(".", maxsplit=1)[0]
    )

    with pysam.AlignmentFile(filename=bam_file, mode="rb", threads=40) as bam_h:
        with pysam.AlignmentFile(
            filename=o_bam_file, mode="wb", threads=40, header=bam_h.header
        ) as o_bam_h:
            for record in tqdm(
                bam_h.fetch(), desc=f"dump_non_multiple_mapping_reads to {o_bam_file}"
            ):
                if record.query_name in qname_set:
                    o_bam_h.write(record)

    pysam.samtools.index("-@", "40", "-b", o_bam_file)


def main(args):
    qname_set = extract_non_multiple_mapping_reads(args.bam)
    dump_non_multiple_mapping_reads(args.bam, qname_set)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("bam")
    main(args=parser.parse_args())



import mappy as mp
import argparse
import re


def calculate_identity_eqx(cigar_str):
    """
    Calculate the identity based on the CIGAR string with `=` and `X` used for match and mismatch respectively.

    Parameters:
    cigar_str (str): The CIGAR string representing the alignment (e.g. '8=2X4=1D5=')

    Returns:
    float: The percentage identity based on the CIGAR string.
    """
    match_count = 0
    mismatch_count = 0
    total_bases = 0

    # Regular expression to match segments in the CIGAR string (e.g. '8=', '2X', '4=', etc.)
    cigar_regex = re.compile(r"(\d+)([=XDI])")

    for match in cigar_regex.finditer(cigar_str):
        length = int(match.group(1))  # Length of this segment
        operation = match.group(2)  # Operation type (=, X, D, etc.)

        if operation == "=":  # Exact match
            match_count += length
            total_bases += length
        elif operation == "X":  # Mismatch
            mismatch_count += length
            total_bases += length
        elif operation == "I":  # Insertion
            total_bases += length
        elif operation == "D":  # Deletion
            total_bases += length
        # For other operations like 'S' (soft clip), 'H' (hard clip), etc., don't count towards total bases.

    if total_bases == 0:
        return 0.0

    # Identity is the fraction of matching bases out of total bases considered in the alignment.
    identity = (match_count / total_bases) * 100
    return identity


def merge_intervals_and_calculate_length(intervals):
    if len(intervals) == 0:
        return [], 0, 0
    # 排序区间，按起始位置排序
    intervals.sort(key=lambda x: x[0])

    merged = []
    total_overlap_length = 0
    prev_start, prev_end = intervals[0]

    for current_start, current_end in intervals[1:]:
        # 如果当前区间与上一个区间重叠或相邻
        if current_start <= prev_end:
            # 计算重叠部分的长度
            overlap_length = min(prev_end, current_end) - current_start
            total_overlap_length += overlap_length

            # 合并区间，更新结束位置
            prev_end = max(prev_end, current_end)
        else:
            # 如果没有重叠，直接添加上一个区间到合并区间
            merged.append([prev_start, prev_end])
            prev_start, prev_end = current_start, current_end

    # 添加最后一个区间
    merged.append([prev_start, prev_end])

    # 计算每个合并后区间的长度
    merged_length = sum([end - start for start, end in merged])

    return merged, merged_length, total_overlap_length


def find_uncovered_part(big_interval, merged_intervals):
    uncovered_parts = []
    big_start, big_end = big_interval

    # 如果没有合并区间或合并区间为空
    if not merged_intervals:
        return [big_interval]

    # 遍历合并区间，计算未被覆盖的部分
    # 1. 检查大区间开始到第一个合并区间之前的部分
    if big_start < merged_intervals[0][0]:
        uncovered_parts.append([big_start, merged_intervals[0][0]])

    # 2. 检查两个合并区间之间的部分
    for i in range(1, len(merged_intervals)):
        prev_end = merged_intervals[i - 1][1]
        current_start = merged_intervals[i][0]

        if prev_end < current_start:
            uncovered_parts.append([prev_end, current_start])

    # 3. 检查最后一个合并区间到大区间结束的部分
    if merged_intervals[-1][1] < big_end:
        uncovered_parts.append([merged_intervals[-1][1], big_end])

    return uncovered_parts


class SingleQueryHits:
    def __init__(
        self, query_name: str, ref_name: str, query_length: int, ref_length: int, hits
    ):
        self.query_name = query_name
        self.ref_name = ref_name
        self.query_length = query_length
        self.ref_length = ref_length
        self.hits = hits

        self.identity = None
        self.query_coverage = None
        self.query_uncovered = None
        self.query_ovlp = None

        self.ref_coverage = None
        self.ref_uncovered = None
        self.ref_ovlp = None

    def fill_field(self):
        cigar_strs = [hit.cigar_str for hit in self.hits]
        self.identity = calculate_identity_eqx(cigar_str="".join(cigar_strs))

        query_start_ends = [[hit.q_st, hit.q_en] for hit in self.hits]
        merged, merged_length, total_overlap_length = (
            merge_intervals_and_calculate_length(query_start_ends)
        )
        self.query_ovlp = total_overlap_length / max(merged_length, 1)
        self.query_uncovered = find_uncovered_part(
            [0, self.query_length], merged_intervals=merged
        )
        self.query_coverage = merged_length / self.query_length

        ref_start_ends = [[hit.r_st, hit.r_en] for hit in self.hits]
        merged, merged_length, total_overlap_length = (
            merge_intervals_and_calculate_length(ref_start_ends)
        )
        self.ref_ovlp = total_overlap_length / max(merged_length, 1)
        self.ref_uncovered = find_uncovered_part(
            [0, self.ref_length], merged_intervals=merged
        )
        self.ref_coverage = merged_length / self.ref_length

    def to_str(self) -> str:
        res = f"""
        -------------------------------------------------------------------
        - query_name: {self.query_name}
        - ref_name: {self.ref_name}
        - identity: {self.identity}
        - 
        - query_coverage: {self.query_coverage}
        - query_uncovered: {self.query_uncovered}
        - query_ovlp: {self.query_ovlp}
        -
        - ref_coverage: {self.ref_coverage}
        - ref_uncovered: {self.ref_uncovered}
        - ref_ovlp: {self.ref_ovlp}
        -------------------------------------------------------------------
        """
        return res


def main(ref_fastx: str, query_fastx: str, preset="map-ont"):
    # 67108864 eqx
    aligner = mp.Aligner(ref_fastx, preset=preset, extra_flags=67108864)
    assert aligner.n_seq == 1
    ref_seq_name = aligner.seq_names[0]
    ref_seq = aligner.seq(ref_seq_name)

    for name, seq, _ in mp.fastx_read(query_fastx):
        hits = list(aligner.map(seq))
        single_query_hits = SingleQueryHits(
            query_name=name,
            ref_name=ref_seq_name,
            query_length=len(seq),
            ref_length=len(ref_seq),
            hits=hits,
        )
        single_query_hits.fill_field()

        print(single_query_hits.to_str())


def main_cli():
    parser = argparse.ArgumentParser(prog="WGA")
    parser.add_argument("ref", type=str, help="ref fastx")
    parser.add_argument("query", type=str, help="query fastx")
    parser.add_argument("--preset", type=str, default="map-ont")
    args = parser.parse_args()

    main(args.ref, args.query, args.preset)


if __name__ == "__main__":
    main_cli()




import os
import argparse
import pathlib
import multiprocessing as mp
import sys
cur_path = pathlib.Path(os.path.abspath(__file__))
cur_dir = cur_path.parent
prev_dir = cur_path.parent.parent
prev_prev_dir = cur_dir.parent.parent.parent
sys.path.append(str(cur_dir))
sys.path.append(str(prev_dir))
sys.path.append(str(prev_prev_dir))

# print(f"cur_dir:{cur_dir}")
# print(f"prev_dir:{prev_dir}")
# print(f"prev_prev_dir:{prev_prev_dir}")
# print(f"sys.path={sys.path}")

import homo_and_str_region_coverage  # noqa: E402
import dna_strand_inconsistency_analysis  # noqa: E402
import asts_analysis  # noqa: E402
import macebell_ratio as macebell_ratio  # noqa: E402


def main(args):
    """
    低Q20 Q30 yield 的问题分析脚本。
    当前进行以下分析：
        1. 是否存在大量正反链不一致
        2. subreads 2 smc 的比对结果是不是很差
        3. 是否大量的 homo+str

    Args:
        args (_type_): 参数
    """
    sbr_bam = args.sbr_bam
    smc_bam = args.smc_bam

    report = "\n\n=================Summary=================\n\n"

    asts_param = {
        "sbr_bam": sbr_bam,
        "smc_bam": smc_bam,
        "rq_range": "0.95:1.0",
        "np_range": "5:1000000"
    }
    asts_param = argparse.Namespace(**asts_param)
    report += asts_analysis.main(asts_param)

    dna_strand_in_consisteny_param = {
        "smc_bam": smc_bam,
        "sbr_bam": sbr_bam,
        "np_thr": 14,
        "rq_thr": 0.95
    }
    dna_strand_in_consisteny_param = argparse.Namespace(
        **dna_strand_in_consisteny_param)
    report += dna_strand_inconsistency_analysis.main(
        dna_strand_in_consisteny_param)

    homo_str_regin_cov_param = {
        "files": [smc_bam],
        "rq_thr": 0.95
    }
    homo_str_regin_cov_param = argparse.Namespace(**homo_str_regin_cov_param)
    report += homo_and_str_region_coverage.main(homo_str_regin_cov_param)

    # ir_ratio_param = {
    #     ""
    # }

    report += macebell_ratio.main(bam_path=smc_bam,
                                  threads=mp.cpu_count())

    print(report)
    pass


def main_cli():
    parser = argparse.ArgumentParser(prog="")
    parser.add_argument("--sbr-bam", dest="sbr_bam")
    parser.add_argument("--smc-bam", dest="smc_bam")

    args = parser.parse_args()
    main(args=args)
    pass


if __name__ == "__main__":
    main_cli()
    pass



import pysam
import re


def extract_channel_ids(fq_file):
    """
    提取 smc.bam 中满足条件的 channel_id 信息
    :param smc_bam: smc.bam 文件路径
    :return: 一个包含符合条件的 channel_id 的集合
    """
    channel_ids = set()
    # 20250604_240901Y0008_Run0001/41790_0_3753

    pattern = r"/(\d+)_"  # 匹配斜杠后的数字，直到下划线

    with pysam.FastxFile(fq_file) as records:
        for record in records:
            channel = re.search(pattern, record.name).group(1)
            channel_ids.add(int(channel))

    return channel_ids


def dump_subreads_by_channel_id(subreads_bam, channel_ids, output_bam):
    """
    将 subreads.bam 中符合 channel_id 条件的记录 dump 到新的 BAM 文件
    :param subreads_bam: subreads.bam 文件路径
    :param channel_ids: 从 smc.bam 中提取的 channel_id 集合
    :param output_bam: 输出的新的 BAM 文件路径
    """
    with pysam.AlignmentFile(
        subreads_bam, "rb", check_sq=False, threads=40
    ) as subreads, pysam.AlignmentFile(
        output_bam, "wb", header=subreads.header, check_sq=False, threads=40
    ) as out_bam:
        for read in subreads.fetch(until_eof=True):
            # 如果该 read 的 channel_id 在指定的集合中，则将其写入新的 BAM 文件

            if read.has_tag("ch") and read.get_tag("ch") in channel_ids:
                out_bam.write(read)


def main():
    """
    主函数：从 smc.bam 中提取符合条件的 channel_id 并从 subreads.bam 中 dump 相应的记录
    :param smc_bam: smc.bam 文件路径
    :param subreads_bam: subreads.bam 文件路径
    :param output_bam: 输出的新的 BAM 文件路径
    """

    anchor_file = "/data/ccs_data/genegle/Barcode-209.fastq"
    target_bam = "/data/ccs_data/genegle/20250604_240901Y0008_Run0001_adapter.smicing.3.smc_all_reads.bam"
    output_bam = "/data/ccs_data/genegle/20250604_240901Y0008_Run0001_adapter.q20.barcode209.smc_all_reads.bam"

    # 提取 smc.bam 中符合条件的 channel_id
    channel_ids = extract_channel_ids(anchor_file)
    print(f"提取到的 channel_id 数量: {len(channel_ids)}")

    # 从 subreads.bam 中 dump 符合条件的记录到 output_bam
    dump_subreads_by_channel_id(target_bam, channel_ids, output_bam)
    print(f"符合条件的记录已保存至 {output_bam}")


if __name__ == "__main__":
    # 示例：调用主函数

    main()




import pysam
import seaborn as sns
import matplotlib.pyplot as plt
from tqdm import tqdm
from collections import Counter
import numpy as np

# 1. 打开 BAM 文件
bam_path = "/data/ccs_data/subreads-dist-smc-dist-discrepency/20250402_240601Y0012_Run0002.smc_all_reads.bam"  # ← 修改为你的 bam 路径
bamfile = pysam.AlignmentFile(bam_path, "rb", check_sq=False, threads=40)

# 2. 提取每个 read 的长度
read_lengths = [read.query_length for read in tqdm(bamfile.fetch(until_eof=True), desc=f"reading {bam_path}")]

read_lengths = np.array(read_lengths)
max_rl = read_lengths.max()

bins = np.arange(0, max_rl + 10 - 1, 10)


# _, bins = np.histogram(read_lengths, bins="auto")
# print("bins:", bins, len(bins))

# print("diffs:", np.diff(bins))
counts = Counter(read_lengths)
counts = sorted(counts.items(), key=lambda x: x[0])
for count in counts:
    if count[0] > 100:
        break
    print(count[0], count[1])
    
# # print(counts)


# print(len(read_lengths))
bamfile.close()

# 3. 可视化分布
sns.set(style="whitegrid")
plt.figure(figsize=(10, 6))
sns.histplot(read_lengths, bins=bins, kde=True, color="skyblue")
plt.title("Read Length Distribution")
plt.xlabel("Read Length")
plt.ylabel("Count")
plt.tight_layout()
plt.savefig("output.png")


import polars as pl
import os


def polars_env_init():
    os.environ["POLARS_FMT_TABLE_ROUNDED_CORNERS"] = "1"
    os.environ["POLARS_FMT_MAX_COLS"] = "100"
    os.environ["POLARS_FMT_MAX_ROWS"] = "300"
    os.environ["POLARS_FMT_STR_LEN"] = "100"


def main():
    # fname = "/data/ccs_data/qunfeng-2025/20250528_250301Y0001_Run0002_called_adapter.polish2.smc_all_reads-metric/20250528_250301Y0001_Run0002_called_adapter.polish2.smc_all_reads.gsmm2_aligned_metric_fact.csv"
    fname = "/data/ccs_data/qunfeng-2025/20250528_250301Y0001_Run0002_called_adapter.polish3.smc_all_reads-metric/20250528_250301Y0001_Run0002_called_adapter.polish3.smc_all_reads.gsmm2_aligned_metric_fact.csv"
    
    data = pl.read_csv(fname, separator="\t")
    data = data.filter(pl.col("rname").str.starts_with(
        "Akkermansia_muciniphila_16s"))
    
    data = data.filter(pl.col("rname").str.starts_with(
        "Akkermansia_muciniphila_16s_3"))
    
    print(len(data))
    print(data.group_by("rname").agg([pl.len()]))
    selected_data = data.select([
        pl.col("qname"), 
        pl.col("identity"),
        pl.col("ins-C"),
        pl.col("homoIns-C"),
        pl.col("del-C"),
        pl.col("homoDel-C"),
        pl.col("ins-G"),
        pl.col("homoIns-G"),
        pl.col("del-G"),
        pl.col("homoDel-G"),
        ])
    print(selected_data)
    print(selected_data.select([pl.col("identity").median()]))
    
    channels = selected_data.select([pl.col("qname").str.split("/").list.last()])
    channels.write_csv(file="/data/ccs_data/qunfeng-2025/expected_channels.csv", include_header=False)

    pass


if __name__ == "__main__":
    polars_env_init()
    main()
    pass



class RangeIdentityCalculator:
    def __init__(self, cigar_tuples):
        self.query_start_pos_and_op = []
        q_start_cursor = 0

        for op, length in cigar_tuples:
            self.query_start_pos_and_op.append((q_start_cursor, (op, length)))
            if op in {0, 1, 4, 7, 8}:  # Match, Ins, SoftClip, Equal, Diff
                q_start_cursor += length

    def range_identity(self, start, end):
        def binary_search(arr, key):
            low, high = 0, len(arr) - 1
            while low <= high:
                mid = (low + high) // 2
                if arr[mid][0] < key:
                    low = mid + 1
                elif arr[mid][0] > key:
                    high = mid - 1
                else:
                    return mid, True
            return low - 1, False

        start_idx, _ = binary_search(self.query_start_pos_and_op, start)
        end_idx, _ = binary_search(self.query_start_pos_and_op, end)

        # Ensure indices are within valid range
        start_idx = max(0, start_idx)
        end_idx = max(0, end_idx)

        res_cigars = []

        if start_idx == end_idx:
            op, length = self.query_start_pos_and_op[start_idx][1]
            start_pos = self.query_start_pos_and_op[start_idx][0]
            span_length = end - start
            overlap_start = max(start, start_pos)
            overlap_end = min(end, start_pos + length)

            if overlap_start >= overlap_end:
                return 0.0  # No valid overlap

            if op in {1, 8}:  # Insertion, Diff
                res_cigars.append((8, overlap_end - overlap_start))  # Treat as Diff
            elif op == 7:  # Equal
                res_cigars.append((7, overlap_end - overlap_start))
        else:
            # Handle the first segment
            start_pos = self.query_start_pos_and_op[start_idx][0]
            op, length = self.query_start_pos_and_op[start_idx][1]
            overlap_start = max(start, start_pos)
            overlap_end = min(end, start_pos + length)
            span_length = overlap_end - overlap_start

            if span_length > 0:
                if op in {1, 8}:  # Insertion, Diff
                    res_cigars.append((8, span_length))  # Treat as Diff
                elif op == 7:  # Equal
                    res_cigars.append((7, span_length))

            # Middle segments
            res_cigars.extend(
                (op, length)
                for _, (op, length) in self.query_start_pos_and_op[start_idx + 1 : end_idx]
            )

            # Handle the last segment
            end_pos = self.query_start_pos_and_op[end_idx][0]
            op, length = self.query_start_pos_and_op[end_idx][1]
            overlap_start = max(start, end_pos)
            overlap_end = min(end, end_pos + length)
            span_length = overlap_end - overlap_start

            if span_length > 0:
                if op in {1, 8}:  # Insertion, Diff
                    res_cigars.append((8, span_length))  # Treat as Diff
                elif op == 7:  # Equal
                    res_cigars.append((7, span_length))

        # Calculate spans and matches
        span_len = sum(length for op, length in res_cigars if op in {1, 7, 8})
        eq_len = sum(length for op, length in res_cigars if op == 7)

        return eq_len / (span_len if span_len > 0 else 1.0)
    
if __name__ == "__main__":
    # Example Usage
    # Example cigar string: [(0, 5), (8, 3), (7, 7), (1, 2)]
    # Operations: MATCH=0, INS=1, DIFF=8, EQUAL=7, etc.
    cigar_tuples = [
        (7, 5),  # eq
        (8, 3), # diff
        (7, 10), 
        (1, 2),
        (7, 6),
        (8, 4)
    ]

    calculator = RangeIdentityCalculator(cigar_tuples)
    identity = calculator.range_identity(0, 5)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(5, 8)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(8, 18)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(3, 7)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(7, 12)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(18, 26)
    print(f"Range Identity: {identity:.2f}")

    cigar_tuples = [
        (7, 5),
        (8, 5)
    ]

    calculator = RangeIdentityCalculator(cigar_tuples)
    identity = calculator.range_identity(0, 3)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(6, 10)
    print(f"Range Identity: {identity:.2f}")

    identity = calculator.range_identity(4, 7)
    print(f"Range Identity: {identity:.2f}")